@@ -50,6 +50,7 @@
 #include "common/optparser.h"
 #include "microdc.h"
 
+static void cmd_retry(int argc, char **argv);
 static void cmd_help(int argc, char **argv);
 static void cmd_exit(int argc, char **argv);
 static void cmd_say(int argc, char **argv);
@@ -192,7 +193,7 @@ command_init(void)
         _("exit"),
         _("Quit the program.\n"));
     add_builtin_command("find", cmd_find, remote_path_completion_generator,
-        _("find [FILE]"),
+        _("find [FILE ...]"),
         _("List files and directories recursively. Assume current directory if FILE is "
           "not specified. Must be browsing a user's files to use this command.\n"));
     add_builtin_command("get", cmd_get, remote_path_completion_generator,
@@ -200,8 +201,10 @@ command_init(void)
         _("Queue file for download. Must be browsing a user's files to use this "
           "command.\n"));
     add_builtin_command("grantslot", cmd_grantslot, user_completion_generator,
-        _("grantslot USER..."),
-        _("Grant a download slot for the specified users.\n"));
+        _("grantslot [USER ...]"),
+        _("Grant a download slot for the specified users, or remove granted slot if the "
+          "user was already granted one. Without arguments, display a list of users with "
+          "granted slots.\n"));
     add_builtin_command("help", cmd_help, command_completion_generator,
         _("help [COMMAND ...]"),
         _("If COMMAND is specified, display help for that command. Otherwise list all "
@@ -213,6 +216,9 @@ command_init(void)
           "\n"
           "Options:\n"
           "  -l, --long    use a long listing format\n"));
+    add_builtin_command("retry", cmd_retry, user_with_queue_completion_generator,
+        _("retry USER ..."),
+        _("Try to connect and download files from the specified users.\n"));
     add_builtin_command("msg", cmd_msg, user_completion_generator,
         _("msg USER MESSAGE..."),
         _("Send a private message to USER. Note that characters such as semicolon "
@@ -226,8 +232,9 @@ command_init(void)
         _("pwd"),
         _("Display user being browsed and current directory.\n"));
     add_builtin_command("queue", cmd_queue, user_with_queue_completion_generator,
-        _("queue USER ..."),
-        _("Display files queued for download from the specified users.\n"));
+        _("queue [USER ...]"),
+        _("Display files queued for download from the specified users. Without "
+          "arguments, display a list of users we have queued files for.\n"));
     add_builtin_command("raw", cmd_raw, NULL,
         _("raw DATA..."),
         _("Send some raw data to the hub. Note that characters such as semicolon "
@@ -261,7 +268,7 @@ command_init(void)
     add_builtin_command("transfers", cmd_transfers, NULL,
         _("transfers"),
         _("Display a list of user connections.\n"));
-    add_builtin_command("unqueue", cmd_unqueue, user_completion_generator,
+    add_builtin_command("unqueue", cmd_unqueue, user_with_queue_completion_generator,
         _("unqueue USER [RANGE]"),
         _("Remove all or a range of queued files for USER. If RANGE is not specified, "
           "remove all files from the queue. Use dash (`-') and comma (`,') in RANGE. "
@@ -351,25 +358,6 @@ default_completion_selector(DCCompletionInfo *ci)
     }
 }
 
-#if 0
-static char *
-replace_substr(const char *str, int sp, int ep, const char *repl)
-{
-    char *out;
-    size_t len;
-    size_t repl_len;
-
-    len = strlen(str);
-    repl_len = strlen(repl);
-    out = xmalloc(sp + strlen(repl) + len - ep + 1);
-    memcpy(out, str, sp);
-    memcpy(out+sp, repl, repl_len);
-    memcpy(out+sp+repl_len, str+ep, len-ep+1);
-
-    return out;
-}
-#endif
-
 static void
 alias_command_completion_selector(DCCompletionInfo *ci)
 {
@@ -565,8 +553,13 @@ executable_completion_generator(DCCompletionInfo *ci)
                     full = catfiles(p, ent->d_name);
                     if (access(full, X_OK) == 0) { /* ignore errors */
                         struct stat sb;
-                        if (stat(full, &sb) >= 0 && S_ISREG(sb.st_mode))
-                            ptrv_append(ci->results, new_completion_entry(ent->d_name, ent->d_name));
+                        if (stat(full, &sb) >= 0 && S_ISREG(sb.st_mode)) {
+                            DCCompletionEntry *ce;
+                            
+                            ce = new_completion_entry(ent->d_name, ent->d_name);
+                            ce->sorting.file_type = DC_TYPE_REG;
+                            ptrv_append(ci->results, ce);
+                        }
                     }
                     /* XXX: use is_directory instead of that stat above */
                     free(full);
@@ -576,6 +569,7 @@ executable_completion_generator(DCCompletionInfo *ci)
         }
     }
 
+    /* local_fs_completion_generator will sort the results. */
     if (*dir_part != '\0' || path_has_cwd) {
         local_fs_completion_generator(ci,DC_CPL_DIR|DC_CPL_EXE|DC_CPL_DOT);
     } else if (ci->results->cur == 0) {
@@ -702,7 +696,6 @@ cmd_shell(int argc, char **argv)
 static void
 cmd_status(int argc, char **argv)
 {
-    char *fmt1;
     HMapIterator it;
     uint32_t c;
 
@@ -737,9 +730,18 @@ cmd_status(int argc, char **argv)
     if (c == 0)
         screen_putf(_("  (none)\n"));
 
-    fmt1 = ngettext("byte", "bytes", my_share_size);
-    screen_putf(_("Share size: %" PRId64" %s (%" PRId64 " MB).\n"),
-                my_share_size, fmt1, my_share_size/(1024*1024));
+    screen_putf(_("Share size: %" PRIu64" %s (%" PRIu64 " MB)\n"),
+                my_share_size,
+                ngettext("byte", "bytes", my_share_size),
+                my_share_size/(1024*1024));
+    screen_putf(_("Bytes received: %" PRIu64 " %s (%" PRIu64 " MB)\n"),
+                bytes_received,
+                ngettext("byte", "bytes", bytes_received),
+                bytes_received/(1024*1024));
+    screen_putf(_("Bytes sent: %" PRIu64 " %s (%" PRIu64 " MB)\n"),
+                bytes_sent,
+                ngettext("byte", "bytes", bytes_sent),
+                bytes_sent/(1024*1024));
 }
 
 static void
@@ -882,14 +884,21 @@ cmd_grantslot(int argc, char **argv)
 {
     uint32_t c;
 
-    if (argc == 1) {
-    	screen_putf(_("Usage: %s USER...\n"), argv[0]);
-	return;
-    }
     if (hub_state != DC_HUB_LOGGED_IN) {
     	screen_putf(_("Not connected.\n"));
 	return;
     }
+    if (argc == 1) {
+        HMapIterator it;
+
+        hmap_iterator(hub_users, &it);
+        while (it.has_next(&it)) {
+            DCUserInfo *ui = it.next(&it);
+            if (ui->slot_granted)
+                screen_putf("%s\n", ui->nick);
+        }
+        return;
+    }
     for (c = 1; c < argc; c++) {
         DCUserInfo *ui;
 
@@ -898,18 +907,24 @@ cmd_grantslot(int argc, char **argv)
     	    screen_putf(_("%s: No such user on this hub\n"), quotearg(argv[c]));
     	    return;
         }
-        ui->slot_granted = true;
+        if (ui->slot_granted) {
+            screen_putf(_("%s has been granted a slot.\n"), ui->nick);
+        } else {
+            screen_putf(_("%s is no longer granted a slot.\n"), ui->nick);
+        }
+        ui->slot_granted = !ui->slot_granted;
     }
 }
 
 /* XXX: move queue.c/browse.c? */
 void
 browse_none(void)
-{	
+{
     /* Clean up previous browse. */
     if (browse_list != NULL) {
-	filelist_free(browse_list);
-	browse_list = NULL;
+        if (!browsing_myself)
+	    filelist_free(browse_list);
+        browse_list = NULL;
 	free(browse_path);
 	browse_path = NULL;
 	free(browse_path_previous);
@@ -979,6 +994,8 @@ cmd_browse(int argc, char **argv)
 	    queued->filename = xstrdup("/MyList.DcLst");
 	    queued->base_path = xstrdup("/");
 	    queued->flag = DC_TF_LIST;
+	    queued->status = DC_QS_QUEUED;
+	    queued->length = UINT64_MAX; /* UINT64_MAX means that the size is unknown */
 	    ptrv_prepend(ui->download_queue, queued);
 	}
     	if (!has_user_conn(ui, DC_DIR_RECEIVE) && ui->conn_count < DC_USER_MAX_CONN)
@@ -1026,16 +1043,16 @@ cmd_pwd(int argc, char **argv)
 static void
 cmd_cd(int argc, char **argv)
 {
-    DCFileList *node;
-    char *path;
-
     if (browse_list == NULL) {
 	screen_putf(_("Not browsing any user.\n"));
 	return;
     }
 
     if (argc == 1) {
-	path = xstrdup("/");
+        free(browse_path_previous);
+        browse_path_previous = browse_path;
+        browse_path = filelist_get_path(browse_list);
+        update_prompt();
     } else if (strcmp(argv[1], "-") == 0) {
 	if (browse_path_previous == NULL) {
 	    warn(_("No previous path.\n"));
@@ -1043,69 +1060,97 @@ cmd_cd(int argc, char **argv)
 	    swap(browse_path, browse_path_previous);
 	    update_prompt();
 	}
-	return;
     } else {
-	path = apply_cwd(argv[1]);
+        bool quoted = false;
+        PtrV *results;
+        DCFileList *basenode;
+        char *basedir;
+
+        results = ptrv_new();
+        if (has_leading_slash(argv[1])) {
+            basenode = browse_list;
+            basedir = "/";
+        } else {
+            basenode = filelist_lookup(browse_list, browse_path);
+            basedir = "";
+        }
+        remote_wildcard_expand(argv[1], &quoted, basedir, basenode, results);
+        if (results->cur >= 1) {
+            char *name = results->buf[0];
+            char *fullname;  
+            DCFileList *node;
+
+            fullname = apply_cwd(name);
+            node = filelist_lookup(browse_list, fullname);
+            if (node != NULL) { /* Technically, this shouldn't fail */
+                if (node->type != DC_TYPE_DIR) {
+                    warn(_("%s: not a directory\n"), quotearg(name));
+                } else {
+                    free(browse_path_previous);
+                    browse_path_previous = browse_path;
+                    browse_path = filelist_get_path(node);
+                    update_prompt();
+                }
+            }
+            free(fullname);
+            ptrv_foreach(results, free);
+        }
+        ptrv_free(results);
     }
 
-    node = filelist_lookup(browse_list, path);
-    if (node == NULL) {
-    	screen_putf(_("%s: No such file or directory\n"), quotearg(argv[1]));
-	free(path);
-	return;
-    }
-    if (node->type != DC_TYPE_DIR) {
-    	screen_putf(_("%s: Not a directory\n"), quotearg(argv[1]));
-	free(path);
-	return;
-    }
-    free(browse_path_previous);
-    browse_path_previous = browse_path;
-    browse_path = filelist_get_path(node);
-    update_prompt();
 }
 
 /* XXX: move queue.c/browse.c? */
 static void
 cmd_find(int argc, char **argv)
 {
-    char *path;
-    DCFileList *node;
+    uint32_t c;
 
     if (browse_list == NULL) {
 	screen_putf(_("Not browsing any user.\n"));
 	return;
     }
+
     if (argc == 1) {
-        path = xstrdup(browse_path);
-    } else {
-	path = apply_cwd(argv[1]);
-    }
+        DCFileList *node;
 
-    node = filelist_lookup(browse_list, path);
-    free(path);
-    if (node == NULL) {
-	screen_putf(_("%s: No such file or directory\n"), quotearg(argv[1]));
-	return;
+        node = filelist_lookup(browse_list, browse_path);
+        /* At the moment, node cannot be NULL here. */
+        if (node != NULL)
+            filelist_list_recursively(node, "");
     }
-    filelist_list_recursively(node, "", 1);
-}
+    for (c = 1; c < argc; c++) {
+	bool quoted = false;
+	PtrV *results;
+	DCFileList *basenode;
+	char *basedir;
+	uint32_t d;
 
-/* XXX: move queue.c/browse.c? */
-static void
-list_directory(const char *path, bool long_mode)
-{
-    char *abspath;
-    DCFileList *node;
+        results = ptrv_new();
+        if (has_leading_slash(argv[c])) {
+            basenode = browse_list;
+            basedir = "/";
+        } else {
+            basenode = filelist_lookup(browse_list, browse_path);
+            basedir = "";
+        }
+        remote_wildcard_expand(argv[c], &quoted, basedir, basenode, results);
+        for (d = 0; d < results->cur; d++) {
+            char *name = results->buf[d];
+            char *fullname;
+            DCFileList *node;
 
-    abspath = apply_cwd(path);
-    node = filelist_lookup(browse_list, abspath);
-    if (node == NULL) {
-	screen_putf(_("%s: No such file or directory\n"), quotearg(path));
-    } else {
-	filelist_list(node, long_mode);
+            fullname = apply_cwd(name);
+            node = filelist_lookup(browse_list, fullname);
+            if (node != NULL) /* Technically, this shouldn't fail */
+                filelist_list_recursively(node, name);
+            free(fullname);
+            free(name);
+        }
+        if (d == 0)
+            screen_putf(_("%s: No such file or directory\n"), quotearg(argv[c]));
+        ptrv_free(results);
     }
-    free(abspath);
 }
 
 /* XXX: move queue.c/browse.c? */
@@ -1114,7 +1159,6 @@ cmd_ls(int argc, char **argv)
 {
     OptDetail long_opts[] = {
 	{ "l", OPTP_NO_ARG, 'l' },
-/*	{ "help", OPTP_NO_ARG, 'h' },*/
 	{ NULL },
     };
     OptParser *p;
@@ -1127,45 +1171,114 @@ cmd_ls(int argc, char **argv)
 	case 'l':
 	    long_mode = true;
 	    break;
-	/*case 'h':
-	    screen_putf("Usage: ls [OPTION]... [FILE]...\n");
-	    screen_putf("  -l          use a long listing format\n");
-	    screen_putf("      --help  display this help and exit\n");
-	    optparser_free(p);
-	    return;*/
 	}
     }
     if (optparser_error(p) != NULL) {
 	screen_putf("%s: %s\n", argv[0], optparser_error(p));
+	optparser_free(p);
 	return;
     }
     if (browse_list == NULL) {
 	screen_putf(_("Not browsing any user.\n"));
+	optparser_free(p);
 	return;
     }
-    if (!optparser_has_next_arg(p))
-	list_directory(browse_path, long_mode);
+    if (!optparser_has_next_arg(p)) {
+        DCFileList *node;
+
+        node = filelist_lookup(browse_list, browse_path);
+        /* At the moment, node cannot be NULL here. */
+        if (node != NULL)
+            filelist_list(node, long_mode);
+    }
     while (optparser_has_next_arg(p)) {
 	char *arg = optparser_next_arg(p);
-	list_directory(arg, long_mode);
+	bool quoted = false;
+	char *basedir;
+	PtrV *results;
+	DCFileList *basenode;
+	uint32_t d;
+
+        results = ptrv_new();
+        if (has_leading_slash(arg)) {
+            basenode = browse_list;
+            basedir = "/";
+        } else {
+            basenode = filelist_lookup(browse_list, browse_path);
+            basedir = "";
+        }
+        remote_wildcard_expand(arg, &quoted, basedir, basenode, results);
+        for (d = 0; d < results->cur; d++) {
+            char *name = results->buf[d];
+            char *fullname;  
+            DCFileList *node;
+                                       
+            fullname = apply_cwd(name);
+            node = filelist_lookup(browse_list, fullname);
+            if (node != NULL) /* Technically, this shouldn't fail */
+                filelist_list(node, long_mode);
+            free(fullname);                       
+            free(name);
+        }
+        if (d == 0)
+            screen_putf(_("%s: No such file or directory\n"), quotearg(arg));
+        ptrv_free(results);
     }
     optparser_free(p);
 }
 
+static void
+cmd_retry(int argc, char **argv)
+{
+    uint32_t c;
+
+    if (hub_state < DC_HUB_LOGGED_IN) {
+    	screen_putf(_("Not connected.\n"));
+	return;
+    }
+
+    for (c = 1; c < argc; c++) {
+        DCUserInfo *ui;
+
+        ui = hmap_get(hub_users, argv[c]);
+        if (ui == NULL) {
+            screen_putf(_("%s: No such user on this hub\n"), quotearg(argv[c]));
+            continue;
+        }
+        if (!has_user_conn(ui, DC_DIR_RECEIVE) && ui->conn_count < DC_USER_MAX_CONN) {
+            hub_connect_user(ui);
+        } else {
+            screen_putf(_("%s: Already connected to user.\n"), quotearg(ui->nick));
+        }
+    }
+}
+
 /* XXX: move to download.c/queue.c/browse.c? */
 static void
 cmd_queue(int argc, char **argv)
 {
     uint32_t d;
 
-    if (argc == 1) {
+    /*if (argc == 1) {
     	screen_putf(_("Usage: %s USER\n"), argv[0]);
 	return;
-    }
+    }*/
     if (hub_state < DC_HUB_LOGGED_IN) {
     	screen_putf(_("Not connected.\n"));
 	return;
     }
+    if (argc == 1) {
+        HMapIterator it;
+
+        hmap_iterator(hub_users, &it);
+        while (it.has_next(&it)) {
+            DCUserInfo *ui = it.next(&it);
+            if (ui->download_queue->cur > 0)
+                screen_putf("%3d %s\n", ui->download_queue->cur, ui->nick);
+        }
+        return;
+    }
+
     for (d = 1; d < argc; d++) {
         uint32_t c;
         DCUserInfo *ui;
@@ -1178,8 +1291,27 @@ cmd_queue(int argc, char **argv)
         screen_putf("%s:\n", quotearg(ui->nick));
         for (c = 0; c < ui->download_queue->cur; c++) {
             DCQueuedFile *queued = ui->download_queue->buf[c];
-            screen_putf(_("%d. [ %s ] %s\n"),
+            char *status;
+
+            switch (queued->status) {
+            case DC_QS_QUEUED:
+                status = "queued";
+                break;
+            case DC_QS_PROCESSING:
+                status = "processing";
+                break;
+            case DC_QS_DONE:
+                status = "done";
+                break;
+            case DC_QS_ERROR:
+            default:
+                status = "error";
+                break;
+            }
+
+            screen_putf("%d. (%s) [%s] %s\n",
                     c+1,
+                    status,
                     quotearg_n(0, queued->base_path),
                     quotearg_n(1, queued->filename + strlen(queued->base_path)));
         }
@@ -1261,6 +1393,42 @@ cmd_unqueue(int argc, char **argv)
 	return;
     }
     foreach_in_range(range, 1, user->download_queue->cur, removed_queued_by_range, user);
+
+    if (has_user_conn(user, DC_DIR_RECEIVE)) {
+        DCUserConn *uc = NULL;
+        uint32_t c;
+
+        for (c = 0; c < user->conn_count; c++) {
+            if (user->conn[c]->dir == DC_DIR_RECEIVE) {
+                uc = user->conn[c];
+                break;
+            }
+        }
+        if (uc->queue_pos < user->download_queue->cur) {
+            uint32_t rank = 0;
+
+            if (user->download_queue->buf[uc->queue_pos] == NULL) {
+                /* Possible actions to take on unqueue of file currently
+                 * being downloaded:
+                 * - abort download, discard file (keep partial)
+                 * - wait until download finishes, and leave as .part
+                 * - wait until download finishes, rename and report
+                 * - not possible to remove such queued file
+                 * The current implementation does (2), which probably
+                 * isn't optimal, but works.
+                 */
+                uc->queued_valid = false;
+            }
+            for (c = 0; c < uc->queue_pos; c++) {
+                if (user->download_queue->buf[c] != NULL)
+                    rank++;
+            }
+            uc->queue_pos = rank;
+            /* uc->queue_pos may be uc->download_queue->cur after
+             * compact_queue_ptrv, but this is OK. */
+        }
+    }
+
     compact_queue_ptrv(user);
 }
 
@@ -1306,7 +1474,7 @@ cmd_who(int argc, char **argv)
                 screen_putf(_("Level: %d\n"), ui->level);
                 screen_putf(_("E-mail: %s\n"), quotearg(IFNULL(ui->email, "")));
                 screen_putf(_("Operator: %d\n"), ui->is_operator);
-                screen_putf(_("Share Size: %" PRId64 " %s (%" PRId64 " MB)\n"), /* " */
+                screen_putf(_("Share Size: %" PRIu64 " %s (%" PRIu64 " MB)\n"), /* " */
                     ui->share_size, fmt1, ui->share_size/(1024*1024));
             }
         }
@@ -1336,7 +1504,7 @@ cmd_who(int argc, char **argv)
 	strbuf_clear(out);
 	strbuf_append(out, nick);
 	strbuf_append_char_n(out, maxlen+1-strlen(nick), ' ');
-	strbuf_appendf(out, "  %7" PRId64 "M", ui->share_size / (1024*1024)); /* " */
+	strbuf_appendf(out, "  %7" PRIu64 "M", ui->share_size / (1024*1024)); /* " */
 	strbuf_append(out, ui->is_operator ? " op" : "   ");
 	if (ui->download_queue->cur > 0)
 	    strbuf_appendf(out, " (%3d)", ui->download_queue->cur);
@@ -1357,35 +1525,30 @@ cmd_transfers(int argc, char **argv)
     char *format;
     HMapIterator it;
     uint32_t maxlen = 0;
-    time_t now = 0;
+    time_t now;
 
     hmap_iterator(user_conns, &it);
     while (it.has_next(&it)) {
     	DCUserConn *uc = it.next(&it);
 	maxlen = max(maxlen, strlen(uc->name));
     }
-
     format = xasprintf("%%-%ds  %%s\n", maxlen);
 
-    if (time(&now) == (time_t) -1) {
+    now = time(NULL);
+    if (now == (time_t) -1)
     	warn(_("Cannot get current time - %s\n"), errstr);
-	return;
-    }
 
     hmap_iterator(user_conns, &it);
     while (it.has_next(&it)) {
     	DCUserConn *uc = it.next(&it);
 	char *status;
 
-    	/*if (!get_user_conn_status(uc))
-	    continue;*/
 	status = user_conn_status_to_string(uc, now);
 	screen_putf(format, quotearg(uc->name), status);
 	free(status);
     }
 
     screen_putf(_("Upload slots: %d/%d  Download slots: %d/unlimited\n"), used_ul_slots, my_ul_slots, used_dl_slots);
-
     free(format);
 }
 
@@ -1622,8 +1785,10 @@ append_download_file(DCUserInfo *ui, DCFileList *node, DCFileList *basenode, uin
 
             queued = xmalloc(sizeof(DCQueuedFile));
             queued->filename = path;
-            queued->base_path = filelist_get_path(basenode);
+            queued->base_path = filelist_get_path_with_trailing_slash(basenode);
             queued->flag = DC_TF_NORMAL;
+            queued->status = DC_QS_QUEUED;
+            queued->length = node->u.reg.size;
             ptrv_append(ui->download_queue, queued);
         }
 
@@ -1639,61 +1804,6 @@ append_download_file(DCUserInfo *ui, DCFileList *node, DCFileList *basenode, uin
     }
 }
 
-#if 0
-static bool
-has_wildcards(const char *str)
-{
-    return strpbrk(str, "*?") != NULL;
-}
-
-/* This function will strip leading slashes in WILD.
- * It assumes that WILD is relative to NODE.
- */
-static void
-wildcard_get(DCFileList *node, char *wild, DCGetData *gd)
-{
-    HMapIterator it;
-    char *next;
-
-    for (; *wild == '/'; wild++);
-    next = strchr(wild, '/');
-
-    if (next != NULL) {
-        *next = '\0';
-        if (has_wildcards(wild)) {
-            hmap_iterator(node->u.dir.children, &it);
-            while (it.has_next(&it)) {
-                node = it.next(&it);
-                /* At this point we can only match directories. */
-                if (node->type == DC_TYPE_DIR) {
-                    if (fnmatch(wild, node->name, 0) == 0)
-                        wildcard_get(node, next+1, gd);
-                }
-            }
-        } else {
-            node = hmap_get(node->u.dir.children, wild);
-            if (node != NULL)
-                wildcard_get(node, next+1, gd);
-        }
-        *next = '/';
-        return;
-    }
-
-    if (has_wildcards(wild)) {
-        hmap_iterator(node->u.dir.children, &it);
-        while (it.has_next(&it)) {
-            node = it.next(&it);
-            if (fnmatch(wild, node->name, 0) == 0)
-                append_download_file(node, gd);
-        }
-    } else {
-        node = hmap_get(node->u.dir.children, wild);
-        if (node != NULL)
-            append_download_file(node, gd);
-    }
-}
-#endif
-
 /* Note: This command assumes that we are browsing some user.
  * If we in the future would allow browse-less getting,
  * i.e. get NICK FILE, then we must here add additional checks
@@ -1721,6 +1831,7 @@ cmd_get(int argc, char **argv)
     for (c = 1; c < argc; c++) {
         DCFileList *basenode;
         bool quoted = false;
+        char *basedir;
         uint64_t byte_count = 0;
         uint32_t file_count = 0;
         PtrV *results;
@@ -1729,19 +1840,30 @@ cmd_get(int argc, char **argv)
         results = ptrv_new();
         if (has_leading_slash(argv[c])) {
             basenode = browse_list;
+            basedir = "/";
         } else {
             basenode = filelist_lookup(browse_list, browse_path);
+            basedir = "";
         }
-        remote_wildcard_expand(argv[c], &quoted, basenode, results);
+        remote_wildcard_expand(argv[c], &quoted, basedir, basenode, results);
         for (d = 0; d < results->cur; d++) {
-            DCFileList *node = results->buf[d];
-            screen_putf(_("Matched %s\n"), quotearg(node->name));
-            append_download_file(browse_user, node, basenode, &file_count, &byte_count);
+            char *name = results->buf[d];
+            char *fullname;  
+            DCFileList *node;
+
+            fullname = apply_cwd(name);
+            node = filelist_lookup(browse_list, fullname);
+            if (node != NULL) { /* Technically, this shouldn't fail */
+                screen_putf(_("Matched %s\n"), quotearg(name));
+                append_download_file(browse_user, node, basenode, &file_count, &byte_count);
+            }
+            free(fullname);                       
+            free(name);
         }
         ptrv_free(results);
 
         if (file_count > 0) {
-            screen_putf(_("Downloading %" PRId64 " %s in %" PRId32 " %s\n"),
+            screen_putf(_("Downloading %" PRIu64 " %s in %" PRIu32 " %s\n"),
                 byte_count, ngettext("byte", "bytes", byte_count),
                 file_count, ngettext("file", "files", file_count));
             dl_some = true;
