@@ -35,28 +35,6 @@
 #include "full-read.h"
 #include "full-write.h"
 
-#ifdef HAVE_FOPENCOOKIE
-static ssize_t byteq_libc_write(void *bq, const char *buf, size_t size);
-
-/* Function table for use with GNU libc's fopencookie.
- */
-static cookie_io_functions_t byteq_libc_io = {
-    NULL, /* read */
-    byteq_libc_write, /* write */
-    NULL, /* seek */
-    NULL, /* close */
-};
-
-/* Some data was written to stream, for example using fprintf.
- */
-static ssize_t
-byteq_libc_write(void *bq, const char *buf, size_t size)
-{
-    byteq_append(bq, (void *) buf, size);
-    return size;
-}
-#endif
-
 /* Arbitrary enlargement of the byteq. This function only guarantees tha
  * the structure will hold one more byte after the call.
  * Note: bq->max may never be 0.
@@ -76,14 +54,6 @@ byteq_new(size_t initial_size)
     ByteQ *bq;
 
     bq = xmalloc(sizeof(ByteQ));
-#ifdef HAVE_FOPENCOOKIE
-    bq->file = fopencookie(bq, "w", byteq_libc_io);
-    if (bq->file == NULL)
-	xalloc_die();
-    if (setvbuf(bq->file, NULL, _IONBF, 0) != 0)
-	xalloc_die();
-#endif
-    
     bq->cur = 0;
     bq->max = MAX(1, initial_size);
     bq->buf = xmalloc(initial_size);
@@ -97,9 +67,6 @@ void
 byteq_free(ByteQ *bq)
 {
     if (bq != NULL) {
-#ifdef HAVE_FOPENCOOKIE
-	fclose(bq->file); /* ignore errors */
-#endif
 	free(bq->buf);
 	free(bq);
     }
@@ -133,9 +100,6 @@ byteq_append(ByteQ *bq, void *data, size_t len)
 int
 byteq_vappendf(ByteQ *bq, const char *format, va_list ap)
 {
-#ifdef HAVE_FOPENCOOKIE
-    return vfprintf(bq->file, format, ap);
-#else
     char *str;
     size_t len;
 
@@ -144,7 +108,6 @@ byteq_vappendf(ByteQ *bq, const char *format, va_list ap)
     byteq_append(bq, str, len);
     free(str);
     return len;
-#endif
 }
 
 /* Append bytes to the byteq, printf style.
