@@ -147,6 +147,12 @@ ptrv_find(PtrV *pv, void *element, comparison_fn_t comparator)
     return -1;
 }
 
+void
+ptrv_sort(PtrV *pv, comparison_fn_t comparator)
+{
+    qsort(pv->buf, pv->cur, sizeof(void *), comparator);
+}
+
 void
 ptrv_insort(PtrV *pv, void *element, comparison_fn_t comparator)
 {
