@@ -35,6 +35,7 @@
 #include "xstrndup.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
 #include "fnmatch.h"		/* Gnulib */
+#include "dirname.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
@@ -60,6 +61,22 @@ struct _DCFileListIterator {
 
 DCFileList *our_filelist = NULL;
 
+static int
+fs_completion_entry_compare(const void *e1, const void *e2)
+{  
+    const DCCompletionEntry *ce1 = *(const DCCompletionEntry **) e1;
+    const DCCompletionEntry *ce2 = *(const DCCompletionEntry **) e2;
+    if (ce1->sorting.file_type == ce2->sorting.file_type) {
+        char *s1 = xasprintf(ce1->display_fmt, ce1->display); /* XXX: this is really slow */
+        char *s2 = xasprintf(ce2->display_fmt, ce2->display); /* XXX: this is really slow */
+        int cmp = strcmp(s1, s2);
+        free(s1);
+        free(s2);
+        return cmp;
+    }
+    return ce1->sorting.file_type - ce2->sorting.file_type;
+}   
+
 static DCFileList *
 new_file_node(const char *name, DCFileType type, DCFileList *parent)
 {
@@ -208,6 +225,23 @@ filelist_lookup(DCFileList *node, const char *filename)
     return NULL;
 }
 
+char *
+filelist_get_path_with_trailing_slash(DCFileList *node)
+{
+    StrBuf *sb;
+
+    sb = strbuf_new();
+    while (node->parent != NULL) {
+	strbuf_prepend(sb, node->name);
+	strbuf_prepend_char(sb, '/');
+    	node = node->parent;
+    }
+    if (node->type == DC_TYPE_DIR)
+        strbuf_append_char(sb, '/');
+
+    return strbuf_free_to_string(sb);
+}
+
 /* Return the path of the node relative to the share root directory.
  */
 char *
@@ -235,20 +269,21 @@ filelist_get_real_path(DCFileList *node)
 {
     char *p1;
     char *p2;
-    
+
     p1 = filelist_get_path(node);
-    p2 = catfiles(share_dir, p1);
+    p2 = catfiles(share_dir, p1+1); /* p1[0] == '/', we don't want that */
     free(p1);
     return p2;
 }
 
-static int
+/*static int
 filelist_completion_compare(const void *i1, const void *i2)
 {
     const DCCompletionEntry *ce1 = *(const DCCompletionEntry **) i1;
     const DCCompletionEntry *ce2 = *(const DCCompletionEntry **) i2;
     return strcmp(ce1->input, ce2->input);
 }
+*/
 
 static int
 file_node_compare(const void *i1, const void *i2)
@@ -286,25 +321,21 @@ get_sorted_file_list(DCFileList *node, uint32_t *out_count)
 }
 
 void
-filelist_list_recursively(DCFileList *node, char *basepath, int skip)
+filelist_list_recursively(DCFileList *node, char *basepath)
 {
     if (node->type == DC_TYPE_DIR) {
-	char *subpath;
 	DCFileList **items;
 	uint32_t c;
 
-        subpath = skip > 0 ? xstrdup(basepath) : catfiles(basepath, node->name);
         items = get_sorted_file_list(node, NULL);
-	for (c = 0; items[c] != NULL; c++)
-            filelist_list_recursively(items[c], subpath, MAX(skip-1, 0));
+	for (c = 0; items[c] != NULL; c++) {
+	    char *path = catfiles(basepath, items[c]->name);
+            filelist_list_recursively(items[c], path);
+            free(path);
+        }
 	free(items);
-	free(subpath);
     } else {
-	char *tmp;
-
-	tmp = catfiles(basepath, node->name);
-	screen_putf("%7" PRIu64 "M %s%s\n", (uint64_t) (node->u.reg.size/(1024*1024)), quotearg(tmp), ""); /* " */
-	free(tmp);
+	screen_putf("%7" PRIu64 "M %s\n", (uint64_t) (node->u.reg.size/(1024*1024)), quotearg(basepath)); /* " */
     }
 }
 
@@ -513,7 +544,7 @@ filelist_to_string(DCFileList *node, StrBuf *sb, int level)
     	strbuf_append_char_n(sb, level-1, '\t');
 
     if (node->type == DC_TYPE_REG) {
-   	strbuf_appendf(sb, "%s|%" PRId64 "\r\n", node->name, node->u.reg.size); /* " joe sh bug */
+   	strbuf_appendf(sb, "%s|%" PRIu64 "\r\n", node->name, node->u.reg.size); /* " joe sh bug */
     } else {
     	HMapIterator it;
 
@@ -598,24 +629,6 @@ filelist_create(const char *basedir)
     return true;
 }
 
-/* Scan for bad components such as '..' in filename.
- */
-static bool
-check_filename(const char *inname)
-{
-    const char *t;
-    const char *t2;
-
-    for (t = inname; (t2 = strchr(t, '/')) != NULL; t = t2+1) {
-    	if (t2-t == 2 && strncmp(t, "..", 2) == 0)
-	    return false;
-    }
-    if (strcmp(t, "..") == 0)
-    	return false;
-
-    return true;
-}
-
 /* Find the physical path of a file that the remote end
  * wants to download from us.
  */
@@ -647,21 +660,21 @@ resolve_upload_file(DCUserInfo *ui, const char *name)
 }
 
 char *
-resolve_download_file(DCUserInfo *ui, const char *inname, const char *base_path)
+resolve_download_file(DCUserInfo *ui, DCQueuedFile *queued)
 {
     char *filename;
+    char *tmp;
 
-    /* Ideally, we would check the user's MyList.DcLst and look
-     * up a DCFileList node.
-     */
-    if (!check_filename(inname))
-    	return NULL;
-    if (strcmp(inname, "/MyList.DcLst") == 0) {
-        filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, ui->nick);
+    if (queued->flag == DC_TF_LIST) {
+        tmp = xasprintf("%s.%s", ui->nick, base_name(queued->filename));
+        filename = catfiles(listing_dir, tmp);
+        free(tmp);
 	ptrv_append(delete_files, xstrdup(filename));
         mkdirs_for_file(filename, true); /* Ignore errors */
     } else {
-        filename = catfiles(download_dir, inname+strlen(base_path));
+        tmp = catfiles(download_dir, queued->filename + strlen(queued->base_path));
+        filename = xasprintf("%s.part", tmp);
+        free(tmp);
         mkdirs_for_file(filename, false); /* Ignore errors */
     }
 
@@ -906,6 +919,7 @@ add_remote_wildcard_result(char *name, DCFileList *node, DCFSCompletionFlags fla
             entry = new_completion_entry_full(input, name, "%s", "%s", true, true);
     }
     if (entry != NULL) {
+        entry->sorting.file_type = node->type;
         ptrv_append(ci->results, entry);
     } else {
         free(input);
@@ -928,8 +942,10 @@ add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInf
 
         input = filename_quote_string(name, dquoted, true);
         if (S_ISDIR(st.st_mode)) {
-            if (flags & DC_CPL_DIR)
+            if (flags & DC_CPL_DIR) {
                 entry = new_completion_entry_full(input, name, "%s/", "%s/", false, true);
+                entry->sorting.file_type = DC_TYPE_DIR;
+            }
         } else if (S_ISLNK(st.st_mode)) {
             if (stat(name, &st) == 0 && S_ISDIR(st.st_mode)) {
                 if (flags & DC_CPL_DIR) {
@@ -940,17 +956,22 @@ add_local_wildcard_result(char *name, DCFSCompletionFlags flags, DCCompletionInf
                         entry->input_single_fmt = "%s/";
                     if (dquoted)
                         input[strlen(input)] = '"';
+                    entry->sorting.file_type = DC_TYPE_DIR;
                 }
             } else {
-                if (flags & DC_CPL_REG)
+                if (flags & DC_CPL_REG) {
                     entry = new_completion_entry_full(input, name, "%s", "%s@", true, true);
+                    entry->sorting.file_type = DC_TYPE_REG;
+                }
             }
         } else if (S_ISREG(st.st_mode)) {
             if ((flags & DC_CPL_REG) || (flags & DC_CPL_EXE)) {
                 if ((access(name, X_OK) == 0) == ((flags & DC_CPL_EXE) != 0)) {
                     entry = new_completion_entry_full(input, name, "%s", "%s", true, true);
+                    entry->sorting.file_type = DC_TYPE_REG;
                 } else if ((flags & DC_CPL_EXE) == 0) {
                     entry = new_completion_entry_full(input, name, "%s", "%s*", true, true);
+                    entry->sorting.file_type = DC_TYPE_REG;
                 }
             }
         }
@@ -995,9 +1016,10 @@ filelist_get_next(DCFileListIterator *it, DCFileList **node, char **name)
 }
 
 void
-remote_wildcard_expand(char *matchpath, bool *quotedptr, DCFileList *basenode, PtrV *results)
+remote_wildcard_expand(char *matchpath, bool *quotedptr, const char *basedir, DCFileList *basenode, PtrV *results)
 {
     char *matchcomp;
+    char *fullpath;
     char *nodename;
     DCFileList *node;
     DCFileListIterator it;
@@ -1007,20 +1029,31 @@ remote_wildcard_expand(char *matchpath, bool *quotedptr, DCFileList *basenode, P
         while (filelist_get_next(&it, &node, &nodename)) {
             if (fnmatch(matchcomp, nodename, FNM_PERIOD) == 0) {
                 if (*matchpath != '\0') {
-                    if (node->type == DC_TYPE_DIR)
-                        remote_wildcard_expand(matchpath, quotedptr, node, results);
+                    if (node->type == DC_TYPE_DIR) {
+                        fullpath = concat_filenames(basedir, nodename);
+                        remote_wildcard_expand(matchpath, quotedptr, fullpath, node, results);
+                        free(fullpath);
+                    }
                 } else {
-                    ptrv_append(results, node);
+                    ptrv_append(results, concat_filenames(basedir, nodename));
+                    /*ptrv_append(results, node);*/
                 }
             }
         }
     } else {
-        node = get_child_node(basenode, matchcomp);
-        if (node != NULL) {
-            if (*matchpath != '\0') {
-                remote_wildcard_expand(matchpath, quotedptr, node, results);
-            } else {
-                ptrv_append(results, node);
+        if (*matchcomp == '\0') {
+            ptrv_append(results, concat_filenames(basedir, matchcomp));
+        } else {
+            node = get_child_node(basenode, matchcomp);
+            if (node != NULL) {
+                if (*matchpath != '\0') {
+                    fullpath = concat_filenames(basedir, matchcomp);
+                    remote_wildcard_expand(matchpath, quotedptr, fullpath, node, results);
+                    free(fullpath);
+                } else {
+                    ptrv_append(results, concat_filenames(basedir, matchcomp));
+                    /*ptrv_append(results, node);*/
+                }
             }
         }
     }
@@ -1053,7 +1086,7 @@ remote_wildcard_complete(char *matchpath, bool *quotedptr, char *basedir, DCFile
         }
     } else {
         fullpath = concat_filenames(basedir, matchcomp);
-        if (*matchpath != '\0') {
+        if (*matchpath != '\0') { /* more components follow after this one? */
             node = get_child_node(basenode, matchcomp);
             if (node != NULL)
                 remote_wildcard_complete(matchpath, quotedptr, fullpath, node, flags, ci, found_wc);
@@ -1063,7 +1096,9 @@ remote_wildcard_complete(char *matchpath, bool *quotedptr, char *basedir, DCFile
              * This behavior is similar to that of GNU readline.
              */
 	    if (found_wc) {
-	        if (strcmp(basedir, "/") != 0) {
+	        if (*matchcomp == '\0') { /* completion word ends in slash */
+                    add_remote_wildcard_result(concat_filenames(basedir, matchcomp), basenode, flags, ci);
+	        } else {
 	            node = get_child_node(basenode, matchcomp);
                     if (node != NULL)
                         add_remote_wildcard_result(concat_filenames(basedir, matchcomp), node, flags, ci);
@@ -1120,8 +1155,11 @@ local_wildcard_complete(char *matchpath, bool *quotedptr, char *basedir, DCFSCom
              * This behavior is similar to that of GNU readline.
              */
 	    if (found_wc) {
-                if (strcmp(basedir, "/") != 0 && lstat(fullpath, &sb) == 0)
+	        if (*matchcomp == '\0') { /* completion word ends in slash */
+	            add_local_wildcard_result(concat_filenames(basedir, matchcomp), flags, ci);
+                } else if (/*strcmp(basedir, "/") != 0 &&*/ lstat(fullpath, &sb) == 0) {
                     add_local_wildcard_result(concat_filenames(basedir, matchcomp), flags, ci);
+                }
             } else {
                 dh = opendir(*basedir == '\0' ? "." : basedir);
                 if (dh != NULL) {
@@ -1187,7 +1225,7 @@ fixup_wildcard_completion_results(DCCompletionInfo *ci)
         }
         free(s1);
  
-        qsort(ci->results->buf, ci->results->cur, sizeof(DCCompletionEntry *), filelist_completion_compare);
+        ptrv_sort(ci->results, fs_completion_entry_compare);
     }
 }
 
@@ -1218,6 +1256,8 @@ remote_fs_completion_generator(DCCompletionInfo *ci, DCFSCompletionFlags flags)
     char *basedir;
     DCFileList *basenode;
 
+    if (browse_list == NULL)
+        return;
     if (has_leading_slash(ci->word_full)) {
         basenode = filelist_lookup(browse_list, "/");
         basedir = "/";
