@@ -38,6 +38,12 @@
 #include "common/strleftcmp.h"
 #include "microdc.h"
 
+ByteQ *hub_recvq = NULL;
+ByteQ *hub_sendq = NULL;
+int hub_socket = -1;
+DCHubState hub_state = DC_HUB_DISCONNECTED;
+HMap *hub_users = NULL;  /* all users on the current hub (UserInfo->nick => UserInfo) */
+
 struct sockaddr_in hub_addr;
 static uint32_t hub_recvq_last = 0;
 char *hub_name = NULL;
@@ -45,7 +51,7 @@ char *hub_name = NULL;
 bool
 send_my_info(void)
 {
-    return hub_putf("$MyINFO $ALL %s %s<%s V:%s,M:%c,H:1/0/0,S:%d>$ $%s%c$%s$%" PRId64 "$|", /* " */
+    return hub_putf("$MyINFO $ALL %s %s<%s V:%s,M:%c,H:1/0/0,S:%d>$ $%s%c$%s$%" PRIu64 "$|", /* " */
 	  my_nick,
 	  my_description, "++", "0.401", is_active ? 'A':'P', my_ul_slots,
 	  my_speed,
@@ -66,15 +72,17 @@ say_user_completion_generator(DCCompletionInfo *ci)
 
 	if (strleftcmp(ci->word, ui->nick) == 0) {
 	    DCCompletionEntry *entry;
-	    entry = new_completion_entry(NULL, NULL); /* XXX: use the _full version */
-	    entry->input = quote_string(ui->nick, ci->word_full[0] == '"', true);
-	    entry->display = xstrdup(ui->nick);
+	    entry = new_completion_entry_full(
+	        quote_string(ui->nick, ci->word_full[0] == '"', true),
+	        xstrdup(ui->nick),
+	        "%s", "%s",
+	        false,
+	        true); /* entry->input and _single_fmt are already quoted */
 	    entry->input_single_fmt = "%s: ";
-	    entry->finalize = false;
-	    entry->quoted = true; /* entry->input and _single_fmt are already quoted */
 	    ptrv_append(ci->results, entry);
 	}
     }
+    ptrv_sort(ci->results, completion_entry_display_compare);
 }
 
 void
@@ -97,6 +105,7 @@ user_completion_generator(DCCompletionInfo *ci)
 	if (strleftcmp(ci->word, ui->nick) == 0)
 	    ptrv_append(ci->results, new_completion_entry(ui->nick, NULL));
     }
+    ptrv_sort(ci->results, completion_entry_display_compare);
 }
 
 void
@@ -111,6 +120,7 @@ user_with_queue_completion_generator(DCCompletionInfo *ci)
 	if (ui->download_queue->cur > 0 && strleftcmp(ci->word, ui->nick) == 0)
 	    ptrv_append(ci->results, new_completion_entry(ui->nick, NULL));
     }
+    ptrv_sort(ci->results, completion_entry_display_compare);
 }
 
 /* Create a new user info structure, representing a user on the hub.
@@ -130,6 +140,7 @@ user_info_new(const char *nick)
     info->share_size = 0;
     info->active_state = DC_ACTIVE_UNKNOWN;
     info->download_queue = ptrv_new();
+    info->slot_granted = false;
     info->refcount = 1;
     info->is_operator = false;
     info->info_quered = false;
@@ -146,6 +157,11 @@ user_info_new(const char *nick)
     if (info->conn[info->conn_count] != NULL)
     	info->conn_count++;
     free(ucname);
+    /*ucname = xasprintf(_("%s|"), nick);
+    info->conn[info->conn_count] = hmap_get(user_conns, ucname);
+    if (info->conn[info->conn_count] != NULL)
+        info->conn_count++;
+    free(ucname);*/
 
     return info;
 }
