@@ -80,10 +80,10 @@ ipc_read(IPC *ipc, void *data, uint32_t datalen)
 }
 
 static bool
-ipc_read_checked(IPC *ipc, void *data, uint32_t datalen)
+ipc_read_checked(IPC *ipc, void *data, uint32_t datalen, uint32_t id)
 {
     if (datalen > ipc->header.size) {
-        ipc->handle_error(ipc, IPC_ERROR_DATA, 0, ipc->userdata);
+        ipc->handle_error(ipc, IPC_ERROR_DATA, id, ipc->userdata);
         return false;
     }
     ipc->header.size -= datalen;
@@ -201,7 +201,7 @@ ipc_put(IPC *ipc, uint32_t id, ...)
 	    }
 	    break;
 	default:
-    	    ipc->handle_error(ipc, IPC_ERROR_DATA, id, ipc->userdata);
+    	    ipc->handle_error(ipc, IPC_ERROR_DATA, -1, ipc->userdata);
 	    result = false;
 	    break;
 	}
@@ -276,46 +276,46 @@ ipc_get(IPC *ipc, uint32_t id, ...)
     for (c = 0; c < ipc->header.arg_count; c++) {
     	IPCArgType type;
 
-	result = ipc_read_checked(ipc, &type, sizeof(IPCArgType));
+	result = ipc_read_checked(ipc, &type, sizeof(IPCArgType), id);
 	if (!result)
 	    break;
 	types[c] = type;
 
 	switch (type) {
 	case IPC_INT64:
-	    result = ipc_read_checked(ipc, &values[c].t_int64, sizeof(values[c].t_int64));
+	    result = ipc_read_checked(ipc, &values[c].t_int64, sizeof(values[c].t_int64), id);
 	    break;
 	case IPC_INT32:
-	    result = ipc_read_checked(ipc, &values[c].t_int32, sizeof(values[c].t_int32));
+	    result = ipc_read_checked(ipc, &values[c].t_int32, sizeof(values[c].t_int32), id);
 	    break;
 	case IPC_INT:
-	    result = ipc_read_checked(ipc, &values[c].t_int, sizeof(values[c].t_int));
+	    result = ipc_read_checked(ipc, &values[c].t_int, sizeof(values[c].t_int), id);
 	    break;
 	case IPC_BOOL:
-	    result = ipc_read_checked(ipc, &values[c].t_bool, sizeof(values[c].t_bool));
+	    result = ipc_read_checked(ipc, &values[c].t_bool, sizeof(values[c].t_bool), id);
 	    break;
 	case IPC_STR:
-	    result = ipc_read_checked(ipc, &values[c].t_str.len, sizeof(values[c].t_str.len));
+	    result = ipc_read_checked(ipc, &values[c].t_str.len, sizeof(values[c].t_str.len), id);
 	    if (result) {
 	    	if (values[c].t_str.len == -1) {
 		    values[c].t_str.chars = NULL;
 		} else {
     		    values[c].t_str.chars = xmalloc(values[c].t_str.len+1);
 		    values[c].t_str.chars[values[c].t_str.len] = '\0';
-		    result = ipc_read_checked(ipc, values[c].t_str.chars, values[c].t_str.len);
+		    result = ipc_read_checked(ipc, values[c].t_str.chars, values[c].t_str.len, id);
 		    if (!result)
 			free(values[c].t_str.chars);
 		}
 	    }
     	    break;
 	default:
-    	    ipc->handle_error(ipc, IPC_ERROR_DATA, 0, ipc->userdata);
+    	    ipc->handle_error(ipc, IPC_ERROR_DATA, id, ipc->userdata);
 	    result = false;
 	    break;
 	}
     }
     if (result && ipc->header.size != 0) {
-    	ipc->handle_error(ipc, IPC_ERROR_DATA, 0, ipc->userdata);
+    	ipc->handle_error(ipc, IPC_ERROR_DATA, id, ipc->userdata);
 	result = false;
     }
 
@@ -329,7 +329,7 @@ ipc_get(IPC *ipc, uint32_t id, ...)
     	IPCArgType type = va_arg(ap, IPCArgType);
 
 	if (type != types[c]) {
-    	    ipc->handle_error(ipc, IPC_ERROR_DATA, 0, ipc->userdata);
+    	    ipc->handle_error(ipc, IPC_ERROR_DATA, id, ipc->userdata);
 	    result = false;
 	    break;
 	}
@@ -355,7 +355,7 @@ ipc_get(IPC *ipc, uint32_t id, ...)
 	}
     }
     if (result && va_arg(ap, IPCArgType) != IPC_END) {
-    	ipc->handle_error(ipc, IPC_ERROR_DATA, 0, ipc->userdata);
+    	ipc->handle_error(ipc, IPC_ERROR_DATA, id, ipc->userdata);
 	result = false;
     }
     va_end(ap);
