@@ -38,7 +38,11 @@
 #include "version-etc.h"	/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
 #include "quotearg.h"		/* Gnulib */
+#include "quote.h"		/* Gnulib */
 #include "sig2str.h"		/* Gnulib */
+#include "xstrndup.h"		/* Gnulib */
+#include "dirname.h"		/* Gnulib */
+#include "minmax.h"		/* Gnulib */
 #include "getline.h"		/* Gnulib/GNU Libc */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
@@ -57,46 +61,7 @@ enum {
 
 #define LISTEN_QUEUE_CONNS 16
 
-/* hub */
-ByteQ *hub_recvq = NULL;
-ByteQ *hub_sendq = NULL;
-int hub_socket = -1;
-DCHubState hub_state = DC_HUB_DISCONNECTED;
-HMap *user_conns = NULL;
-
-/* hub_users holds all users on the current hub.
- *   UserInfo->nick => UserInfo
- */
-HMap *hub_users = NULL;
-
-/* pending_userinfo holds all users we want to communicate with,
- * either us connecting to them or they connecting to us.
- *   UserInfo->nick => UserInfo
- */
-HMap *pending_userinfo = NULL; 
-
-struct sockaddr_in local_addr;
-struct in_addr force_listen_addr = { INADDR_NONE };
-    
-/* settings/variables */
-int used_ul_slots = 0;
-int used_dl_slots = 0;
-uint32_t my_ul_slots;
-char *my_nick;
-char *my_tag;
-char *my_description;
-char *my_speed;
-char *my_email;
-char *share_dir = NULL;
-char *download_dir;
-char *listing_dir;
-char *my_password;
-bool is_active;
-uint64_t my_share_size = 0;
-uint32_t display_flags = ~(DC_DF_DEBUG); /* All flags except debug set */
-uint32_t log_flags = ~(DC_DF_DEBUG);
-
-/* interactive state */
+/* interactive state (move to transfer.c/browse.c) */
 DCFileList *browse_list = NULL;
 DCUserInfo *browse_user = NULL;
 char *browse_path = NULL;
@@ -104,6 +69,8 @@ char *browse_path_previous = NULL;
 bool browsing_myself = false;
 
 /* misc */
+uint64_t bytes_received = 0;
+uint64_t bytes_sent = 0;
 uint16_t listen_port = 0;
 bool running = true;
 fd_set read_fds;
@@ -112,6 +79,11 @@ PtrV *delete_files = NULL;
 PtrV *delete_dirs = NULL;
 static PtrV *search_udpmsg_out;	/* pending outgoing search results */
 pid_t shell_child = -1;
+struct sockaddr_in local_addr;
+HMap *user_conns = NULL;
+HMap *pending_userinfo = NULL;	/* users we want to communicate with, either
+                                 * us connecting to them or they connecting
+                                 * to us. (UserInfo->nick => UserInfo) */
 
 /* XXX: change to ("Copyright %s %d Oskar Liljeblad", gettext("(C)"), 2005) */
 const char version_etc_copyright[] = "Copyright (C) 2004, 2005 Oskar Liljeblad";
@@ -181,6 +153,11 @@ validate_nick(DCUserConn *uc, const char *nick)
 	    uc->info->conn[uc->info->conn_count++] = uc;
 	    hmap_remove(pending_userinfo, ui->nick);
 	    user_info_free(ui);
+	    /*hmap_remove(user_conns, uc->name);
+	    ptrv_append(user_conn_unknown_free, uc->name);
+	    new_name = xasprintf("%s|", uc->info->nick);
+	    uc->name = new_name;
+	    hmap_put(user_conns, uc->name, uc);*/
 	    return true;
 	}
         warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
@@ -233,7 +210,6 @@ validate_direction(DCUserConn *uc, DCTransferDirection dir)
     uc->name = newname;
     uc->dir = dir;
     hmap_put(user_conns, uc->name, uc);
-
     /* Don't check for a free slot here. We will do that later. */
 
     /* Make new connection if we want to download but couldn't. */
@@ -259,14 +235,107 @@ handle_user_ipc_error(IPC *ipc, IPCError error, int result, void *userdata)
     	warn_socket_error(result, false, _("user process"));
 	break;
     case IPC_ERROR_DATA:
-    	//XXX: assert not reached
-    	warn(_("Logical communication error with user process.\n"));
+    	warn(_("Logical communication error with user process (during message %d).\n"), result);
     	break;
     }
     hmap_remove(user_conns, uc->name);
     user_disconnect(uc);
 }
 
+static void
+handle_ended_upload(DCUserConn *uc, bool success)
+{
+    uint64_t len = uc->transfer_pos - uc->transfer_start;
+
+    bytes_sent += len;
+    if (uc->occupied_slot) {
+        /* Only display "Upload failed" if "Starting upload" was
+         * displayed previously.
+         */
+        flag_putf(DC_DF_UPLOAD, _("%s: Upload of %s %s. %" PRIu64 " %s.\n"),
+            quotearg_n(0, uc->info->nick),
+            quote_n(1, base_name(uc->transfer_file)),
+            success ? _("succeeded") : _("failed"),
+            len,
+            ngettext("byte transferred", "bytes transferred", len));
+        used_ul_slots--;
+        uc->occupied_slot = false;
+    }
+    free(uc->transfer_file);
+    uc->transfer_file = NULL;
+    uc->transfer_start = 0;
+    uc->transfer_pos = 0;
+    uc->transferring = false;
+}
+
+static void
+handle_ended_download(DCUserConn *uc, bool success, const char *reason)
+{
+    uint64_t len = uc->transfer_pos - uc->transfer_start;
+
+    bytes_received += len;
+    if (uc->occupied_slot) {
+        used_dl_slots--;
+        uc->occupied_slot = false;
+    }
+    /* If we removed from the queued item during download,
+     * queue_pos will point outside the queue. */
+    if (uc->queued_valid) {
+        DCQueuedFile *queued;
+
+        queued = uc->info->download_queue->buf[uc->queue_pos];
+        uc->queue_pos++;
+        uc->queued_valid = false;
+        if (success) {
+            queued->status = DC_QS_DONE;
+            if (queued->flag == DC_TF_LIST) {
+                if (browse_user != NULL && strcmp(browse_user->nick, uc->info->nick) == 0 && browse_list == NULL) {
+                    char *filename;
+
+                    screen_putf(_("Now browsing %s.\n"), quotearg(browse_user->nick));
+                    filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, browse_user->nick);
+                    browse_list = filelist_open(filename);
+                    browse_path = xstrdup("/");
+                    browse_path_previous = NULL;
+                    free(filename);
+                    update_prompt();
+                }
+            } else {
+                char *local_file = resolve_download_file(uc->info, queued);
+                char *final_file = xstrndup(local_file, strlen(local_file)-5);
+                if (safe_rename(local_file, final_file) != 0) {
+                    warn(_("%s: Cannot rename file to %s - %s\n"), quotearg_n(0, local_file), quote_n(1, final_file), errstr);
+                    reason = "cannot rename file"; /* XXX: would like a more elaborate error message here perhaps? */
+                    queued->status = DC_QS_ERROR;
+                    success = false;
+                }
+                free(final_file);
+                free(local_file);
+            }
+        } else {
+            queued->status = DC_QS_ERROR;
+        }
+        flag_putf(DC_DF_DOWNLOAD, _("%s: Download of %s %s (%s). %" PRIu64 " %s.\n"),
+            quotearg_n(0, uc->info->nick),
+            quote_n(1, base_name(uc->transfer_file)),
+            success ? _("succeeded") : _("failed"),
+            reason, len,
+            ngettext("byte transferred", "bytes transferred", len));
+    } else {
+        flag_putf(DC_DF_DOWNLOAD, _("%s: Download of %s %s but unqueued (%s). %" PRIu64 " %s.\n"),
+            quotearg_n(0, uc->info->nick),
+            quote_n(1, base_name(uc->transfer_file)),
+            success ? _("succeeded") : _("failed"),
+            reason, len,
+            ngettext("byte transferred", "bytes transferred", len));
+    }
+    free(uc->transfer_file);
+    uc->transfer_file = NULL;
+    uc->transfer_start = 0;
+    uc->transfer_pos = 0;
+    uc->transferring = false;
+}
+
 static bool
 handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
 {
@@ -278,38 +347,22 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
     	uint32_t flag;
 
     	ipc_get(ipc, id, IPC_INT32, &flag, IPC_STR, &msg, IPC_END);
-	/* Don't quote what the user prints */
-    	flag_putf(flag, _("User %s: %s"), quotearg(uc->name), msg);
+    	flag_putf(flag, _("User %s: %s"), quotearg(uc->name), msg); /* msg already quoted */
 	free(msg);
     	return true;
     }
-    case DC_MSG_NEXT_DOWNLOAD: {
-    	DCQueuedFile *queued;
-
-    	ipc_get(ipc, id, IPC_END);
-        if (uc->info->download_queue->cur == 0)
-            return ipc_put(ipc, id, IPC_STR, NULL, IPC_STR, NULL, IPC_INT32, 0, IPC_END);
-        queued = uc->info->download_queue->buf[0];
-	if (!ipc_put(ipc, id, IPC_STR, queued->filename, IPC_STR, queued->base_path, IPC_INT32, queued->flag, IPC_END))
-	    return false;
-        ptrv_remove_range(uc->info->download_queue, 0, 1);
-        free_queued_file(queued);
-    	return true;
-    }
     case DC_MSG_WANT_DOWNLOAD: {
     	bool reply;
 
     	ipc_get(ipc, id, IPC_END);
-	if (has_user_conn(uc->info, DC_DIR_RECEIVE))
-    	    reply = false;
-	else
-    	    reply = uc->info->download_queue->cur > 0;
+    	reply = !has_user_conn(uc->info, DC_DIR_RECEIVE)
+    	        && (uc->queue_pos < uc->info->download_queue->cur);
 	return ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
     }
     case DC_MSG_VALIDATE_DIR: {
     	DCTransferDirection dir;
 	bool reply;
-    
+
     	if (!ipc_get(ipc, id, IPC_INT, &dir, IPC_END))
 	    return false;
     	reply = validate_direction(uc, dir);
@@ -342,92 +395,100 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
     }
     case DC_MSG_GET_MY_NICK: {
         return ipc_get(ipc, id, IPC_END) && ipc_put(ipc, id, IPC_STR, my_nick, IPC_END);
-    };
-    case DC_MSG_RESOLVE_FILE: {
-    	char *inname;
-	char *inbase;
-	char *outname;
-	bool result;
-
-    	if (!ipc_get(ipc, id, IPC_STR, &inname, IPC_STR, &inbase, IPC_END))
-	    return false;
-	if (uc->dir == DC_DIR_SEND) {
-            outname = resolve_upload_file(uc->info, inname);
-	} else { /* uc->dir == DC_DIR_RECEIVE */
-    	    outname = resolve_download_file(uc->info, inname, inbase);
-	}
-	free(inname);
-	free(inbase);
-	result = ipc_put(ipc, id, IPC_STR, outname, IPC_END);
-	free(outname);
-        return result;
     }
-    case DC_MSG_TRANSFER_STARTING: {
-    	bool reply;
-	char *filename;
-	uint64_t filesize;
-
-    	assert(!uc->occupied_slot);
-    	if (!ipc_get(ipc, id, IPC_INT64, &filesize, IPC_STR, &filename, IPC_END))
-	    return false;
-    	if (uc->dir == DC_DIR_SEND) {
-	    if (used_ul_slots >= my_ul_slots && !uc->info->slot_granted) {
-		reply = false;
-	    	free(filename);
-	    } else {
-	    	reply = true;
-    	    	used_ul_slots++;
-	    }
-	} else { /* uc->dir must be DC_DIR_RECV */
-	    reply = true;
-	    used_dl_slots++;
-	}
-	if (reply) {
-	    uc->transfer_pos = 0;
-	    uc->occupied_slot = true;
-	    free(uc->transfer_file);
-    	    uc->transfer_file = filename;
-	    uc->transfer_len = filesize;
-	    if (time(&uc->transfer_start) == (time_t) -1) {
-	    	warn(_("Cannot get current time - %s\n"), errstr);
-		uc->transfer_start = 0;
-	    }
-	}
-	reply = ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
-	return reply;
-    }
-    case DC_MSG_TRANSFER_FINISHED: {
-	uint32_t flag;
-
-    	assert(uc->occupied_slot);
-    	if (!ipc_get(ipc, id, IPC_INT32, &flag, IPC_END))
-	    return false;
-    	if (uc->dir == DC_DIR_SEND) {
-	    used_ul_slots--;
-	} else { /* uc->dir must be DC_DIR_RECV */
-	    used_dl_slots--;
-	}
-	free(uc->transfer_file);
-	uc->transfer_file = NULL;
-	uc->occupied_slot = false;
-	if (flag == DC_TF_LIST) {
-	    if (browse_user != NULL && strcmp(browse_user->nick, uc->info->nick) == 0 && browse_list == NULL) {
-		char *filename;
-
-		screen_putf(_("Now browsing %s.\n"), quotearg(browse_user->nick));
-		filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, browse_user->nick);
-		browse_list = filelist_open(filename);
-		browse_path = xstrdup("/");
-		browse_path_previous = NULL;
-		free(filename);
-		update_prompt();
-	    }
-	}
-    	return true;
-    }
-    case DC_MSG_STATUS: {
+    case DC_MSG_TRANSFER_STATUS: {
     	return ipc_get(uc->ipc, id, IPC_INT64, &uc->transfer_pos, IPC_END);
     }
+    case DC_MSG_TRANSFER_START: {
+    	if (!ipc_get(ipc, id, IPC_INT64, &uc->transfer_start, IPC_INT64, &uc->transfer_total, IPC_END))
+	    return false;
+        if (uc->transfer_start != 0) {
+	    flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " of %" PRIu64 " %s).\n"),
+	            quotearg_n(0, uc->info->nick),
+                    uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
+                    quote_n(1, base_name(uc->transfer_file)),
+                    uc->transfer_total - uc->transfer_start,
+                    uc->transfer_total,
+                    ngettext("byte", "bytes", uc->transfer_total));
+        } else {
+	    flag_putf(DC_DF_CONNECTIONS, _("%s: Starting %s of %s (%" PRIu64 " %s).\n"),
+                    quotearg_n(0, uc->info->nick),
+                    uc->dir == DC_DIR_SEND ? _("upload") : _("download"),
+                    quote_n(1, base_name(uc->transfer_file)),
+                    uc->transfer_total,
+                    ngettext("byte", "bytes", uc->transfer_total));
+        }
+        uc->transferring = true;
+        uc->transfer_pos = uc->transfer_start;
+        uc->transfer_time = time(NULL);
+        if (uc->transfer_time == (time_t) -1)
+            warn(_("Cannot get current time - %s\n"), errstr);
+	return true;
+    }
+    case DC_MSG_CHECK_DOWNLOAD: {
+    	if (!ipc_get(ipc, id, IPC_END))
+    	    return false;
+
+        for (; uc->queue_pos < uc->info->download_queue->cur; uc->queue_pos++) {
+    	    DCQueuedFile *queued;
+    	    char *local_file;
+    	    bool result;
+
+            queued = uc->info->download_queue->buf[uc->queue_pos];
+            if (queued->status != DC_QS_DONE) {
+                local_file = resolve_download_file(uc->info, queued);
+                uc->queued_valid = true;
+                uc->transfer_file = xstrdup(queued->filename);
+                uc->occupied_slot = true;
+                queued->status = DC_QS_PROCESSING;
+                used_dl_slots++;
+                result = ipc_put(ipc, id, IPC_STR, local_file, IPC_STR, uc->transfer_file, IPC_INT64, queued->length, IPC_END);
+                free(local_file);
+		return result;
+            }
+        }
+        return ipc_put(ipc, id, IPC_STR, NULL, IPC_STR, NULL, IPC_INT64, (uint64_t) 0, IPC_END);
+    }
+    case DC_MSG_DOWNLOAD_ENDED: {
+        bool success;
+        char *reason;
+
+    	if (!ipc_get(ipc, id, IPC_BOOL, &success, IPC_STR, &reason, IPC_END))
+	    return false;
+        handle_ended_download(uc, success, reason);
+        free(reason);
+    	return true;
+    }
+    case DC_MSG_CHECK_UPLOAD: {
+        char *remote_file;
+        char *local_file;
+        bool result;
+
+    	if (!ipc_get(ipc, id, IPC_STR, &remote_file, IPC_END))
+	    return false;
+        if (used_ul_slots >= my_ul_slots && !uc->info->slot_granted) {
+            uc->transfer_file = NULL;
+            uc->occupied_slot = false;
+            local_file = NULL;
+            free(remote_file);
+        } else {
+            uc->transfer_file = remote_file;
+            uc->occupied_slot = true;
+            used_ul_slots++;
+            local_file = resolve_upload_file(uc->info, remote_file);
+        }
+        result = ipc_put(ipc, id, IPC_BOOL, uc->occupied_slot, IPC_STR, local_file, IPC_END);
+        free(local_file);
+        return result;
+    };
+    case DC_MSG_UPLOAD_ENDED: {
+        bool success;
+
+        if (!ipc_get(ipc, id, IPC_BOOL, &success, IPC_END))
+            return false;
+        handle_ended_upload(uc, success);
+        return true;
+    };
     case DC_MSG_DISCONNECT:
         /* The DISCONNECT message will not be sent from users, only to users. */
     default:
@@ -438,7 +499,112 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
     }
 }
 
-// XXX: move this to user!?
+DCUserConn *
+user_connection_new(struct sockaddr_in *addr, int user_socket)
+{
+    DCUserConn *uc;
+    int main_sockets[2];
+    pid_t pid;
+
+    if (socketpair(AF_LOCAL, SOCK_STREAM, 0, main_sockets) < 0) {
+    	warn(_("Cannot create socket pair - %s\n"), errstr);
+	return NULL;
+    }
+
+    pid = fork();
+    if (pid < 0) {
+    	warn(_("Cannot create process - %s\n"), errstr);
+	if (close(main_sockets[0]) < 0)
+	    warn(_("Cannot close socket - %s\n"), errstr);
+	if (close(main_sockets[1]) < 0)
+	    warn(_("Cannot close socket - %s\n"), errstr);
+	return NULL;
+    }
+    if (pid == 0)
+	user_main(main_sockets, addr, user_socket);
+
+    if (close(main_sockets[1]) < 0)
+    	warn(_("Cannot close socket - %s\n"), errstr);
+    if (user_socket >= 0 && close(user_socket) < 0)
+    	warn(_("Cannot close socket - %s\n"), errstr);
+
+    uc = xmalloc(sizeof(DCUserConn));
+    uc->pid = pid;
+    uc->main_socket = main_sockets[0];
+    uc->info = NULL;
+    uc->occupied_slot = 0;
+    uc->dir = DC_DIR_UNKNOWN;
+    uc->transfer_file = NULL;
+    uc->transfer_start = 0;
+    uc->transfer_pos = 0;
+    uc->transfer_total = 0;
+    uc->transfer_time = (time_t) -1;
+    uc->transferring = false;
+    uc->queue_pos = 0;
+    uc->queued_valid = false;
+    /* uc->we_connected = (user_socket < 0); */
+    add_user_conn(uc);
+    FD_SET(uc->main_socket, &read_fds);
+
+    return uc;
+}
+
+void
+user_disconnect(DCUserConn *uc)
+{
+    if (uc->occupied_slot) { /* could also check that uc->transfer_file != NULL */
+        if (uc->dir == DC_DIR_SEND) {
+            handle_ended_upload(uc, false);
+        } else { /* uc->dir == DC_DIR_RECV */
+            handle_ended_download(uc, false, "connection terminated prematurely");
+        }
+    }
+
+    flag_putf(DC_DF_CONNECTIONS, _("User %s: Shutting down user connection.\n"), quotearg(uc->name)); /* XXX: move where? */
+    ipc_free(uc->ipc);
+    if (uc->info != NULL) {
+    	int c;
+	for (c = 0; c < uc->info->conn_count; c++) {
+	    if (uc->info->conn[c] == uc)
+		break;
+	}
+	assert(c < uc->info->conn_count);
+	uc->info->conn[c] = NULL;
+    	uc->info->conn_count--;
+	for (; c < uc->info->conn_count; c++)
+	    uc->info->conn[c] = uc->info->conn[c+1];
+	user_info_free(uc->info);
+    }
+    FD_CLR(uc->main_socket, &read_fds);
+    FD_CLR(uc->main_socket, &write_fds);
+    if (close(uc->main_socket) < 0)
+	warn(_("Cannot close socket - %s\n"), errstr);
+    free(uc);
+}
+
+char *
+user_conn_status_to_string(DCUserConn *uc, time_t now)
+{
+    char *status;
+
+    if (uc->transferring) {
+    	int percent = (uc->transfer_pos * 100) / MAX(uc->transfer_total, 1);
+	int rate = 0;
+
+	if (now != (time_t) -1 && uc->transfer_time != (time_t) -1)
+	    rate = (uc->transfer_pos-uc->transfer_start)/1024/MAX(now-uc->transfer_time, 1);
+	status = xasprintf(
+	        uc->dir == DC_DIR_RECEIVE
+		    ? _("Downloading %3d%% (at %5d kb/s) %s")
+		    : _("Uploading   %3d%% (at %5d kb/s) %s"),
+                percent, rate, base_name(uc->transfer_file));
+    } else {
+	status = xasprintf(_("Idle"));
+    }
+
+    return status;
+}
+
 void
 add_user_conn(DCUserConn *uc)
 {
@@ -489,6 +655,7 @@ transfer_completion_generator(DCCompletionInfo *ci)
 	if (strleftcmp(ci->word, uc->name) == 0)
 	    ptrv_append(ci->results, new_completion_entry(uc->name, NULL));
     }
+    ptrv_sort(ci->results, completion_entry_display_compare);
 }
 
 /* This is the actual signal handler, registered in main_init with
@@ -618,8 +785,10 @@ set_share_dir(const char *dir)
     if (hub_state >= DC_HUB_LOGGED_IN && !send_my_info())
     	return false;
     if (dir != NULL) {
-        char *fmt1 = ngettext("byte", "bytes", my_share_size);
-	screen_putf(_("Sharing %" PRId64 " %s (%" PRId64 " MB) totally\n"), my_share_size, fmt1, my_share_size/(1024*1024));
+	screen_putf(_("Sharing %" PRIu64 " %s (%" PRIu64 " MB) totally\n"),
+	    my_share_size,
+	    ngettext("byte", "bytes", my_share_size),
+	    my_share_size/(1024*1024));
     }
     return true;
 }
