@@ -98,6 +98,9 @@ typedef enum {
     DC_HUB_LOGGED_IN,		/* Correctly logged in */
 } DCHubState;
 
+/* The order of fields in DCFileType is important for sorting
+ * completion entries as well.
+ */
 typedef enum {
     DC_TYPE_DIR,
     DC_TYPE_REG,
@@ -105,16 +108,17 @@ typedef enum {
 
 typedef enum {
     DC_MSG_SCREEN_PUT,
-    DC_MSG_NEXT_DOWNLOAD,
     DC_MSG_WANT_DOWNLOAD,
     DC_MSG_VALIDATE_DIR,
     DC_MSG_VALIDATE_NICK,
     DC_MSG_GET_MY_NICK,
-    DC_MSG_RESOLVE_FILE, /* for download or upload */
-    DC_MSG_STATUS,
     DC_MSG_DISCONNECT,
-    DC_MSG_TRANSFER_STARTING,
-    DC_MSG_TRANSFER_FINISHED,
+    DC_MSG_CHECK_DOWNLOAD,	/* get information on next download, allocate slot. */
+    DC_MSG_CHECK_UPLOAD,	/* check that upload is allowed, allocate slot. */
+    DC_MSG_UPLOAD_ENDED,	/* free slot. */
+    DC_MSG_DOWNLOAD_ENDED,	/* free slot and mark download as done or failed. */
+    DC_MSG_TRANSFER_START,
+    DC_MSG_TRANSFER_STATUS,
 } DCUserMsgId;
 
 typedef enum {
@@ -129,6 +133,13 @@ typedef enum {
     DC_SEARCH_CHECKSUM,     /* DC++: TTH at the moment */
 } DCSearchDataType;
 
+typedef enum {
+    DC_QS_QUEUED,	/* also meaning, "not touched" */
+    DC_QS_PROCESSING,	/* currently being downloaded or about to be downloaded */
+    DC_QS_DONE,		/* succesfully downloaded (XXX or already complete) */
+    DC_QS_ERROR,	/* error occured during download */
+} DCQueuedStatus;
+
 typedef struct _DCUserConn DCUserConn;
 typedef struct _DCUserInfo DCUserInfo;
 typedef struct _DCFileList DCFileList;
@@ -232,6 +243,13 @@ struct _DCCompletionEntry {
      * finalize.
      */
     bool quoted;
+    /* These fields are used for sorting purposes.
+     * This is currently only used by remote and local fs
+     * completion generators.
+     */
+    union {
+        DCFileType file_type;
+    } sorting;
 };
 
 struct _DCFileList {
@@ -254,6 +272,8 @@ struct _DCQueuedFile {
     char *filename;  /* XXX: should make this relative, not absolute */
     char *base_path; /* XXX: so that catfiles(base_path, filename) works. */
     DCTransferFlag flag;
+    DCQueuedStatus status;
+    uint64_t length;
 };
 
 struct _DCUserInfo {
@@ -264,33 +284,33 @@ struct _DCUserInfo {
     char *email;
     uint64_t share_size;
     DCActiveState active_state;
-    PtrV *download_queue;   // XXX LList?
+    PtrV *download_queue;
     uint16_t slot_granted;
     uint32_t refcount;
     bool info_quered;
     bool is_operator;
 
-    // Valid elements: conn[0..conn_count-1]
-    DCUserConn *conn[DC_USER_MAX_CONN];
-    // Valid range: 0 <= conn_count <= DC_USER_MAX_CONN
-    int conn_count;
+    DCUserConn *conn[DC_USER_MAX_CONN];	/* valid elements: conn[0..conn_count-1] */
+    int conn_count; /* valid range: 0 <= conn_count <= DC_USER_MAX_CONN */
 };
 
 struct _DCUserConn {
     char *name;     	/* key of connection */
-    // If info!=NULL then info->conn must contain this DCUserConn.
-    DCUserInfo *info;
+    DCUserInfo *info;	/* if info!=NULL then info->conn must contain this DCUserConn. */
     DCTransferDirection dir;
     pid_t pid;
     int main_socket;
     IPC *ipc;
-    bool occupied_slot;
-    //bool we_connected;
-
-    char *transfer_file;    /* valid iff occupied_slot */
-    uint64_t transfer_pos;  /* valid iff occupied_slot */
-    uint64_t transfer_len;  /* valid if(!) occupied_slot */
-    time_t transfer_start;  /* valid iff occupied_slot */
+    bool occupied_slot; /* true if used_*_slots were increased for this user connection */
+    uint32_t queue_pos;
+    bool queued_valid;		/* true unless the QueuedFile by queue_pos has been removed */
+    char *transfer_file;
+    bool transferring;
+    uint64_t transfer_start;	/* the byte where the transfer started */
+    uint64_t transfer_pos;	/* current byte position */
+    uint64_t transfer_total;	/* total number of bytes to transfer */
+    time_t transfer_time;
+    /*bool we_connected;*/
 };
 
 struct _DCVariable {
@@ -369,7 +389,6 @@ void user_with_queue_completion_generator(DCCompletionInfo *ci);
 /* command.c */
 void command_execute(const char *line);
 void default_completion_selector(DCCompletionInfo *ci);
-/*bool append_download_dir(DCUserInfo *ui, DCFileList *node, char *base_path, uint64_t *bytes, uint32_t *files) XXX move transfer.c? */
 void update_prompt(void);
 void command_init(void);
 void command_finish(void);
@@ -395,15 +414,15 @@ void get_file_dir_part(const char *word,char **dir_part,const char **file_part);
 void fill_completion_info(DCCompletionInfo *ci);
 char *quote_string(const char *str, bool dquotes, bool finalize);
 char *filename_quote_string(const char *str, bool dquotes, bool finalize);
+int completion_entry_display_compare(const void *e1, const void *e2);
+int completion_entry_int_userdata_compare(const void *e1, const void *e2);
 
 /* huffman.c */
 char *huffman_decode(const uint8_t *data, uint32_t data_size, uint32_t *out_size);
 char *huffman_encode(const uint8_t *data, uint32_t data_size, uint32_t *out_size);
 
 /* user.c */
-DCUserConn *user_connection_new(struct sockaddr_in *addr, int socket);
-void user_disconnect(DCUserConn *uc);
-char *user_conn_status_to_string(DCUserConn *uc, time_t now);
+void user_main(int main_sockets[2], struct sockaddr_in *addr, int sock);
 
 /* main.c */
 void add_user_conn(DCUserConn *uc);
@@ -419,6 +438,11 @@ void warn_file_error(int res, bool write, const char *filename);
 void warn_socket_error(int res, bool write, const char *subject);
 void add_search_result(struct sockaddr_in *addr, char *results, uint32_t resultlen);
 void free_queued_file(DCQueuedFile *qf); /* XXX: move transfer.c? */
+DCUserConn *user_connection_new(struct sockaddr_in *addr, int socket);
+void user_disconnect(DCUserConn *uc);
+char *user_conn_status_to_string(DCUserConn *uc, time_t now);
+extern uint64_t bytes_received;
+extern uint64_t bytes_sent;
 
 /* fs.c */
 void filelist_free(DCFileList *fl);
@@ -426,8 +450,9 @@ void filelist_list(DCFileList *fl, bool longmode);
 DCFileList *filelist_lookup(DCFileList *node, const char *filename);
 DCFileList *filelist_open(const char *filename);
 char *filelist_get_path(DCFileList *node);
+char *filelist_get_path_with_trailing_slash(DCFileList *node);
 char *resolve_upload_file(DCUserInfo *ui, const char *name);
-char *resolve_download_file(DCUserInfo *ui, const char *inname, const char *base_path);
+char *resolve_download_file(DCUserInfo *ui, DCQueuedFile *queued);
 bool filelist_create(const char *dir);
 char *translate_local_to_remote(const char *localname);
 char *translate_remote_to_local(const char *remotename);
@@ -438,8 +463,8 @@ void remote_path_completion_generator(DCCompletionInfo *ci);
 void remote_dir_completion_generator(DCCompletionInfo *ci);
 char *apply_cwd(const char *path); /* XXX: move transfer.c? */
 extern DCFileList *our_filelist;
-void filelist_list_recursively(DCFileList *node, char *basepath, int skip);
-void remote_wildcard_expand(char *mtachpath, bool *quotedptr, DCFileList *basenode, PtrV *results);
+void filelist_list_recursively(DCFileList *node, char *basepath);
+void remote_wildcard_expand(char *matchpath, bool *quotedptr, const char *basedir, DCFileList *basenode, PtrV *results);
 bool has_leading_slash(const char *str);
     
 /* connection.c */
@@ -448,6 +473,7 @@ char *escape_message(char *str);
 void dump_command(const char *header, const char *buf, size_t len);
 
 /* util.c */
+int safe_rename(const char *oldpath, const char *newpath);
 int ilog10(uint64_t c);
 int mkdirs_for_file(char *filename, bool deletedirs);
 char *catfiles(const char *p1, const char *p2);
@@ -465,9 +491,6 @@ char *in_addr_str(struct in_addr addr);
 bool parse_ip_and_port(char *source, struct sockaddr_in *addr, uint16_t defport);
 #define xquotestr(s) 		(quotearg_alloc((s), SIZE_MAX, NULL))
 #define quotearg_mem(m,l)	(quotearg_n_style_mem(0, escape_quoting_style, (m), (l)))
-#if 0
-char *xconfstr(const char *name);
-#endif
 char *join_strings(char **strs, int count, char mid);
 PtrV *wordwrap(const char *str, size_t len, size_t first_width, size_t other_width);
 
