@@ -370,12 +370,7 @@ screen_finish(void)
 	}
 	free(path);
 
-    	//rl_basic_word_break_characters = ?
-	//rl_completer_word_break_characters = ?
-    	//rl_completion_display_matches_hook = ?
-          //rl_attempted_completion_function = NULL;
-	//rl_char_is_quoted_p = NULL;
-    	  //rl_pre_input_hook = NULL;
+        rl_pre_input_hook = NULL;
 
 	warn_writer = default_warn_writer;
         screen_state = SCREEN_UNINITIALIZED;
@@ -412,27 +407,6 @@ fill_completion_info(DCCompletionInfo *ci)
     ci->word_index = get_word_index(ci->line, ci->ws);
 }
 
-#if 0
-/* Strip a trailing unescaped double quote from STR, if there is any.
- * You can say that this "opens up" STR.
- */
-static void
-strip_trailing_quote(char *str)
-{
-    bool quoted = false;
-
-    for (; str[0] != '\0'; str++) {
-        if (str[0] == '"') {
-            if (str[1] == '\0' && quoted)
-                str[0] = '\0';
-            quoted = !quoted;
-        } else if (str[0] == '\\' && str[1] != '\0') {
-            str++;
-        }
-    }
-}
-#endif
-
 /* Return escaped text to display for the specified completion entry.
  */
 static char *
@@ -586,15 +560,6 @@ screen_prepare(void)
 
 	rl_readline_name = PACKAGE;
 	rl_add_defun(PACKAGE "-complete", completion_readline, '\t');
-	  //rl_attempted_completion_function = attempted_completion;
-	  //rl_completer_quote_characters = "\"";
-	/* rl_basic_word_break_characters = ... */
-	  //rl_special_prefixes = "\"";
- 	  //rl_completer_word_break_characters = " \t\n;\"";  // also \"
-	/* rl_filename_quote_characters = " \t\n\\\"'>;|&()*?[]~!"; */
-	/* rl_filename_quoting_function = quote_argument; */
-	/* rl_filename_dequoting_function = dequote_argument_rl; */
-          //rl_char_is_quoted_p = char_is_quoted;
     	rl_pre_input_hook = fix_winch;
 
         using_history();
@@ -665,3 +630,16 @@ sorted_list_completion_generator(const char *base, PtrV *results, void *items,
         }
     }
 }
+
+int
+completion_entry_display_compare(const void *e1, const void *e2)
+{
+    const DCCompletionEntry *ce1 = *(const DCCompletionEntry **) e1;
+    const DCCompletionEntry *ce2 = *(const DCCompletionEntry **) e2;
+    char *s1 = xasprintf(ce1->display_fmt, ce1->display); /* XXX: this is really slow */
+    char *s2 = xasprintf(ce2->display_fmt, ce2->display); /* XXX: this is really slow */
+    int cmp = strcmp(s1, s2);
+    free(s1);
+    free(s2);
+    return cmp;
+}
