@@ -243,7 +243,7 @@ append_result(DCFileList *node, DCUserInfo *ui, struct sockaddr_in *addr)
 
     strbuf_appendf(sb, "$SR %s %s", my_nick, rpath);
     if (node->type == DC_TYPE_REG)
-    	strbuf_appendf(sb, "\x05%" PRId64, node->u.reg.size);
+    	strbuf_appendf(sb, "\x05%" PRIu64, node->u.reg.size);
     strbuf_appendf(sb, " %d/%d\x05%s (%s)",
     	    free_slots, my_ul_slots, hub_name,
 	    sockaddr_in_str(&hub_addr));
