@@ -56,24 +56,15 @@
 typedef struct _DCUserConnLocal DCUserConnLocal;
 
 struct _DCUserConnLocal {
-    char *user_nick/* = NULL*/;
-    char *base_path/* = NULL*/;
-    char *file_name/* = NULL*/;	/* file we wish to download, in local dirlist namespace */
-    //char *file_name/*UD*//* = NULL*/;	/* part of DownloadState */
-    char *localfile/* = NULL*/; 		/* part of DownloadState */
-    DCTransferFlag file_flag;	/* part of DownloadState */
     int main_socket/* = -1*/;
     int signal_pipe[2]/* = { -1, -1 }*/;
+    char *user_nick/* = NULL*/;
     uint32_t user_recvq_last/* = 0*/;
     ByteQ *user_recvq;
     ByteQ *user_sendq;
     int user_socket/* = -1*/;
-    int transfer_file/* = -1*/;
-    uint64_t data_size/* = 0*/;	/* only useful when receiving files */
+    uint64_t data_size/* = 0*/;		/* only useful when receiving files */
     uint16_t dir_rand;
-    uint64_t file_pos/* = 0*/;
-    uint64_t transfer_pos/* = 0*/;
-    uint64_t file_size/* = 0*/;
     bool we_connected;
     DCTransferDirection our_dir;
     DCUserState user_state/* = DC_USER_CONNECT*/;
@@ -81,6 +72,26 @@ struct _DCUserConnLocal {
     fd_set user_read_fds;
     fd_set user_write_fds;
     IPC *ipc;
+
+    //union {
+      //struct {
+        char *share_file;	/* complete filename in shared file namespace. */
+        char *local_file;	/* complete filename in local physical file namespace. */
+        int transfer_fd;	/* file descriptor for opened local_file */
+        uint64_t file_pos;	/* how much of local_file that has been written */
+        uint64_t file_size;	/* the final size of local_file */
+        uint64_t transfer_pos;	/* how much that has been read (but not necessarily written yet) */
+        bool local_exists;	/* does local_file exist already? */
+      //} dl;
+      //struct {
+        //char *share_file;	/* complete filename in shared file namespace. */
+        //char *local_file;	/* complete filename in local physical file namespace. */
+        //int fd;			/* file descriptor for opened local_file */
+        //uint64_t file_pos;	/* how much of local_file that has been read */
+        //uint64_t file_size;	/* the size of local_file */
+        //uint64_t transfer_pos;	/* how much of what's been read that has been sent to remote */
+      //} ul;
+    //};
 };
 
 static DCUserConnLocal *cur_ucl;
@@ -104,7 +115,7 @@ check_state(DCUserConnLocal *ucl, char *buf, DCUserState state)
 static bool
 send_user_status(DCUserConnLocal *ucl, uint64_t pos)
 {
-    return ipc_put(ucl->ipc, DC_MSG_STATUS, IPC_INT64, pos, IPC_END);
+    return ipc_put(ucl->ipc, DC_MSG_TRANSFER_STATUS, IPC_INT64, pos, IPC_END);
 }
 
 static bool
@@ -118,9 +129,6 @@ handle_main_msg(IPC *ipc, uint32_t id, void *userdata)
 	ipc_put(ipc, id, IPC_END); /* Ignore errors */
     	ucl->user_running = false;
 	return false;
-    //case DC_MSG_STATUS:
-    //	ipc_get(ipc, id, IPC_END);
-    //	return ipc_put(ipc, id, IPC_INT64, transfer_pos, IPC_END);
     default:
         warn(_("Received unknown message %d from main process, shutting down.\n"), id);
 	ucl->user_running = false;
@@ -142,7 +150,7 @@ handle_main_ipc_error(IPC *ipc, IPCError error, int result, void *userdata)
     	warn_socket_error(result, false, _("main process"));
 	break;
     case IPC_ERROR_DATA:
-    	warn(_("Logical communication error with main process.\n"));
+    	warn(_("Logical communication error with main process (during message %d).\n"), result);
     	break;
     }
     ucl->user_running = false;
@@ -155,7 +163,7 @@ user_screen_writer(DCDisplayFlag flag, const char *format, va_list args)
     char *msg;
 
     msg = xvasprintf(format, args);
-    ipc_put(ucl->ipc, DC_MSG_SCREEN_PUT, IPC_INT32, flag, IPC_STR, msg, IPC_END); /* ignore errors */
+    ipc_put(ucl->ipc, DC_MSG_SCREEN_PUT, IPC_INT32, (int32_t) flag, IPC_STR, msg, IPC_END); /* ignore errors */
     free(msg);
 }
 
@@ -195,40 +203,6 @@ user_putf(DCUserConnLocal *ucl, const char *format, ...)
     return true;
 }
 
-static bool
-download_next_file(DCUserConnLocal *ucl)
-{
-    char *remote_file;
-    char *dl_file;
-    char *dl_base;
-
-    if (!ipc_put(ucl->ipc, DC_MSG_NEXT_DOWNLOAD, IPC_END))
-    	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_NEXT_DOWNLOAD, IPC_STR, &dl_file, IPC_STR, &dl_base, IPC_INT32, &ucl->file_flag, IPC_END))
-    	return false;
-    if (dl_file == NULL) {
-        flag_putf(DC_DF_CONNECTIONS, _("No more files to download.\n"));/*XXX:move main.c?*/
-        ucl->user_running = false;
-	return false;
-    }
-
-    free(ucl->file_name/*DL*/);
-    free(ucl->base_path);
-    ucl->file_name/*DL*/ = dl_file;
-    free(ucl->localfile);
-    ucl->localfile = NULL;
-    ucl->base_path = dl_base;
-    remote_file = translate_local_to_remote(dl_file);
-    if (!user_putf(ucl, "$Get %s$1|", remote_file)) {
-        free(remote_file);
-	return false;
-    }
-    free(remote_file);
-    ucl->user_state = DC_USER_FILE_LENGTH;
-
-    return true;
-}
-
 static bool
 wants_to_download(DCUserConnLocal *ucl, bool *reply)
 {
@@ -283,183 +257,274 @@ nick_validate(DCUserConnLocal *ucl, const char *str)
     return true;
 }
 
-static bool
-open_download_file(DCUserConnLocal *ucl)
+static void
+end_download(DCUserConnLocal *ucl, bool success, const char *reason_fmt, ...)
 {
-    bool reply;
+    va_list args;
+    char *reason = NULL;
 
-    if (!ipc_put(ucl->ipc, DC_MSG_RESOLVE_FILE, IPC_STR, ucl->file_name/*DL*/, IPC_STR, ucl->base_path, IPC_END))
-    	return false; /* maybe free file_name here? */
-    if (!ipc_get(ucl->ipc, DC_MSG_RESOLVE_FILE, IPC_STR, &ucl->localfile, IPC_END))
-    	return false; /* maybe free file_name here? */
-    if (ucl->localfile == NULL) {
-	ucl->user_running = false;
-	return false; /* maybe free file_name here? */
+    if (ucl->transfer_fd/*DL*/ >= 0) {
+        /* Inability to close the downloaded file is a serious error. */
+        if (close(ucl->transfer_fd/*DL*/) < 0) {
+            warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->local_file/*DL*/), errstr);
+            if (success) { /* don't let this error obscure a previous error */
+                reason = xstrdup("local error");
+                success = false;
+            }
+        }
+        ucl->transfer_fd/*DL*/ = -1;
     }
 
-    if (ucl->file_flag == DC_TF_NORMAL) {
-        char *tmp = xasprintf(_("%s.part"), ucl->localfile);
-        free(ucl->localfile);
-        ucl->localfile = tmp;
-    }
+    va_start(args, reason_fmt);
+    if (reason == NULL)
+        reason = xvasprintf(reason_fmt, args);
+    ipc_put(ucl->ipc, DC_MSG_DOWNLOAD_ENDED, IPC_BOOL, success, IPC_STR, reason, IPC_END);
+    free(reason);
+    va_end(args);
 
-    ucl->transfer_file/*DL*/ = open/*64*/(ucl->localfile, O_CREAT|O_EXCL|O_WRONLY, 0644);
-    if (ucl->transfer_file/*DL*/ < 0) {
-	warn(_("%s: Cannot open file for writing - %s\n"), quotearg(ucl->localfile), errstr);
-	//free(ucl->localfile); ucl->localfile = NULL;
-	//free(ucl->file_name/*DL*/); ucl->file_name/*DL*/ = NULL;
-	/* XXX: should really try to download file later instead... */
-        //if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_FAILED/*DL*/, IPC_INT32, ucl->file_flag, IPC_END))
-        //   return;
-	return download_next_file(ucl);
-    }
-
-    if (ucl->file_size == 0) {
-	warn(_("File %s is empty, not starting download.\n"), quotearg(ucl->file_name/*DL*/));
-	if (close(ucl->transfer_file/*DL*/) < 0)
-	    warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*DL*/), errstr);
-	ucl->transfer_file/*DL*/ = -1;
-    } else {
-        if (!user_putf(ucl, "$Send|"))
-	    return false; /* maybe free file_name here? */
-        /* XXX: this could be DC_DF_DOWNLOAD as well - but that is only for completed downloads!? */
-        /* XXX: this could move to main:DC_MSG_TRANSFER_STARTING */
-        flag_putf(DC_DF_CONNECTIONS, _("Starting download for file %s, %" PRIu64 " bytes\n"), quotearg(ucl->file_name/*DL*/), ucl->file_size);
-    }
-
-    /* Let the main process think we are starting the download even if
-     * the file is empty and we are not going to download it.
-     * XXX: The code in main.c should be rewritten to handle this better.
-     */
-    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_STARTING/*DL*/, IPC_INT64, ucl->file_size, IPC_STR, ucl->file_name/*DL*/, IPC_END))
-	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_TRANSFER_STARTING/*DL*/, IPC_BOOL, &reply, IPC_END))
-    	return false;
-    if (!reply) {
-	ucl->user_running = false;
-    	return false;
-    }
-
-    if (ucl->file_size == 0) {
-        if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_FINISHED/*DL*/, IPC_INT32, ucl->file_flag, IPC_END))
-           return false;
-	return download_next_file(ucl);
-    }
-    /* Places to go from here (assuming file_size != 0):
-     *   user_running becomes false
-     *     User process termination. user_disconnect deals with used_*_slots.
-     *   data received until file complete
-     *     Will issue DC_MSG_TRANSFER_FINISHED prior to calling
-     *     download_next_file.
-     */
-    ucl->file_pos = 0;
-    ucl->transfer_pos = 0;
-    ucl->data_size/*DL*/ = ucl->file_size;
-    ucl->user_state = DC_USER_DATA_RECV;
-    return true;
+    free(ucl->share_file/*DL*/);
+    ucl->share_file/*DL*/ = NULL; /* if "user terminated" calls this function, then this is not necessary */
+    free(ucl->local_file/*DL*/);
+    ucl->local_file/*DL*/ = NULL; /* if "user terminated" calls this function, then this is not necessary */
 }
 
-/* Returns user_running (!) */
-static bool
+/* This is called when the next file should be downloaded.
+ * It will send $Get to the user.
+ */
+static void
+download_next_file(DCUserConnLocal *ucl)
+{
+    char *share_file;
+    char *local_file;
+    char *remote_file;
+    uint64_t resume_pos;
+    uint64_t file_size;
+    struct stat sb;
+
+    if (!ucl->user_running)
+        return;
+    if (!ipc_put(ucl->ipc, DC_MSG_CHECK_DOWNLOAD, IPC_END))
+    	return;
+    if (!ipc_get(ucl->ipc, DC_MSG_CHECK_DOWNLOAD, IPC_STR, &local_file, IPC_STR, &share_file, IPC_INT64, &file_size, IPC_END))
+    	return;
+    if (local_file == NULL) {
+        flag_putf(DC_DF_CONNECTIONS, _("No more files to download.\n"));/*XXX:move main.c?*/
+        ucl->user_running = false;
+	return;
+    }
+
+    ucl->share_file/*DL*/ = share_file;
+    ucl->local_file/*DL*/ = local_file;
+
+    /* XXX: what if file without ".part" exists and is complete? */
+    /* Check if file already exists, and if it need to be resumed. */
+    if (lstat(local_file, &sb) != 0) {
+        if (errno != ENOENT) {
+            warn(_("%s: Cannot get file status: %s\n"), quotearg(local_file), errstr);
+            end_download(ucl, false, "local error");
+            download_next_file(ucl);
+            return;
+        }
+        ucl->local_exists = false;
+        resume_pos = 0;
+    } else {
+        if (!S_ISREG(sb.st_mode)) {
+            warn(_("%s: File exists and is not a regular file\n"), quotearg(local_file));
+            end_download(ucl, false, "local error");
+            download_next_file(ucl);
+            return;
+        }
+        ucl->local_exists = true;
+        resume_pos = sb.st_size;
+    }
+
+    remote_file = translate_local_to_remote(share_file);
+    if (!user_putf(ucl, "$Get %s$%" PRIu64 "|", remote_file, resume_pos+1)) { /* " */
+        end_download(ucl, false, "communication error");
+        free(remote_file);
+	return;
+    }
+    free(remote_file);
+    ucl->user_state = DC_USER_FILE_LENGTH;
+    ucl->file_size = file_size;
+    ucl->file_pos = resume_pos;
+    ucl->transfer_pos = resume_pos;
+}
+
+static void
+open_download_file(DCUserConnLocal *ucl, uint64_t file_size)
+{
+    if (ucl->file_size == UINT64_MAX) {
+        /* A file size of UINT64_MAX means that we did not know the size
+         * of the file in advance which is only true for file list files.
+         */
+        ucl->file_size = file_size;
+    } else if (file_size < ucl->file_size) {
+        end_download(ucl, false,
+            "remote file is smaller than local (expected %" PRIu64 ", got %" PRIu64 " %s)",
+            ucl->file_size, file_size, ngettext("byte", "bytes", file_size));
+        /* We're probably left in a weird state here. It is possible that
+         * some clients (at least DC++ 0.700) won't let us download more
+         * files from here.
+         */
+        download_next_file(ucl);
+        return;
+    }
+
+    if (ucl->local_exists) { /* Resuming */
+        ucl->transfer_fd/*DL*/ = open/*64*/(ucl->local_file/*DL*/, O_CREAT|O_WRONLY, 0644);
+    } else {
+        ucl->transfer_fd/*DL*/ = open/*64*/(ucl->local_file/*DL*/, O_CREAT|O_EXCL|O_WRONLY, 0644);
+    }
+    if (ucl->transfer_fd/*DL*/ < 0) {
+        warn(_("%s: Cannot open file for writing - %s\n"), quotearg(ucl->local_file/*DL*/), errstr);
+        end_download(ucl, false, "local error");
+        download_next_file(ucl);
+        return;
+    }
+    if (ucl->file_pos != 0 && lseek(ucl->transfer_fd/*DL*/, ucl->file_pos, SEEK_SET) < 0) {
+        warn(_("%s: Cannot seek to resume position - %s\n"), quotearg(ucl->local_file/*DL*/), errstr);
+        end_download(ucl, false, "local error");
+        download_next_file(ucl);
+        return;
+    }
+
+    if (!user_putf(ucl, "$Send|")) {
+        end_download(ucl, false, "communication error");
+        return;
+    }
+
+    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_START, IPC_INT64, ucl->file_pos, IPC_INT64, ucl->file_size, IPC_END)) {
+        end_download(ucl, false, "internal communication error");
+	return;
+    }
+
+    ucl->data_size/*DL*/ = ucl->file_size - ucl->file_pos;
+    if (ucl->data_size/*DL*/ == 0) {
+        end_download(ucl, true, "no data to transfer");
+        download_next_file(ucl);
+        return;
+    }
+    ucl->user_state = DC_USER_DATA_RECV;
+
+    /* Places to go from here
+     *   user_running becomes false
+     *     User process termination. user_disconnect deals with used_*_slots.
+     *   data received until file complete (possible no data)
+     *     Will issue DC_MSG_DOWNLOAD_ENDED prior to calling
+     *     download_next_file.
+     */
+}
+
+static void
+end_upload(DCUserConnLocal *ucl, bool success)
+{
+    if (ucl->transfer_fd/*DL*/ >= 0) {
+        close(ucl->transfer_fd/*DL*/);  /* Ignore errors */
+        ucl->transfer_fd/*DL*/ = -1;
+    }
+    free(ucl->share_file/*UL*/);
+    ucl->share_file/*UL*/ = NULL;
+    free(ucl->local_file/*UL*/);
+    ucl->local_file/*UL*/ = NULL;
+    ipc_put(ucl->ipc, DC_MSG_UPLOAD_ENDED, IPC_BOOL, success, IPC_END);
+}
+
+static void
 open_upload_file(DCUserConnLocal *ucl, const char *str, uint64_t offset)
 {
     struct stat st;
-    char *newname;
-    bool reply;
-    char *uploadfile;
+    bool may_upload;
+    char *share_file;
+    char *local_file;
 
-    newname = translate_remote_to_local(str);
-    if (!ipc_put(ucl->ipc, DC_MSG_RESOLVE_FILE, IPC_STR, newname, IPC_STR, NULL, IPC_END)) {
-    	free(newname);
-	return false;
+    share_file = translate_remote_to_local(str);
+    if (!ipc_put(ucl->ipc, DC_MSG_CHECK_UPLOAD, IPC_STR, share_file, IPC_END)) {
+        free(share_file);
+        return;
     }
-    if (!ipc_get(ucl->ipc, DC_MSG_RESOLVE_FILE, IPC_STR, &uploadfile, IPC_END))
-    	return false;
-
-    if (uploadfile == NULL) {
-        warn(_("Cannot resolve upload file `%s' - invalid filename or no sharedir\n"), quotearg(newname));/*XXX:move main.c */
-	free(newname);
-	if (!user_putf(ucl, "$Error File Not Available|"))
-	    return false;
-	ucl->user_running = false;
-	return false;
+    if (!ipc_get(ucl->ipc, DC_MSG_CHECK_UPLOAD, IPC_BOOL, &may_upload, IPC_STR, &local_file, IPC_END)) {
+        free(share_file);
+        return;
+    }
+    if (!may_upload) {
+    	user_putf(ucl, "$MaxedOut|");
+	free(share_file);
+    	return;
     }
-    free(newname);
 
-    free(ucl->file_name/*UL*/);
-    ucl->file_name/*UL*/ = uploadfile;
+    ucl->share_file/*UL*/ = share_file;
+    ucl->local_file/*UL*/ = local_file;
 
-    /* Get file status and check offset. */
-    if (stat(ucl->file_name/*UL*/, &st) < 0) {
-        warn(_("%s: Cannot get file status - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
-	if (!user_putf(ucl, "$Error File Not Available|"))
-	    return false;
-        return true;
+    if (ucl->local_file/*UL*/ == NULL) {
+        flag_putf(DC_DF_CONNECTIONS, _("%s: No such shared file\n"), quotearg(ucl->share_file/*UL*/));
+	user_putf(ucl, "$Error File Not Available|");
+	end_upload(ucl, false);
+	return;
+    }
+    if (stat(ucl->local_file/*UL*/, &st) < 0) {
+        flag_putf(DC_DF_CONNECTIONS, _("%s: Cannot get file status - %s\n"), quotearg(ucl->local_file/*UL*/), errstr);
+        user_putf(ucl, "$Error File Not Available|");
+        end_upload(ucl, false);
+        return;
     }
     if (offset > st.st_size) {
-	warn(_("%s: Offset %" PRId64 " out of range\n"), quotearg(ucl->file_name/*UL*/), offset);
-	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
-	if (!user_putf(ucl, "$Error Offset out of range|"))
-	    return false;
-        return true;
+	flag_putf(DC_DF_CONNECTIONS, _("%s: Resume offset %" PRIu64 " outside file\n"), quotearg(ucl->local_file/*UL*/), offset);
+	user_putf(ucl, "$Error Offset out of range|");
+	end_upload(ucl, false);
+	return;
     }
-
-    /* Open file and seek to desired position. */
-    ucl->transfer_file/*UL*/ = open/*64*/(ucl->file_name/*UL*/, O_RDONLY);
-    if (ucl->transfer_file/*UL*/ < 0) {
-	warn(_("%s: Cannot open file for reading - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
-	if (!user_putf(ucl, "$Error File Not Available|"))
-	    return false;
-        return true;
+    ucl->transfer_fd/*UL*/ = open/*64*/(ucl->local_file/*UL*/, O_RDONLY);
+    if (ucl->transfer_fd/*UL*/ < 0) {
+	flag_putf(DC_DF_CONNECTIONS, _("%s: Cannot open file for reading - %s\n"), quotearg(ucl->local_file/*UL*/), errstr);
+	user_putf(ucl, "$Error File Not Available|");
+	end_upload(ucl, false);
+	return;
     }
-    if (offset != 0 && lseek/*64*/(ucl->transfer_file/*UL*/, offset, SEEK_SET) < 0) {
-    	warn(_("%s: Cannot seek in file - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-	if (!user_putf(ucl, "$Error File Not Available|"))
-	    return false;
-	if (close(ucl->transfer_file/*UL*/) < 0)
-	    warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-	ucl->transfer_file/*UL*/ = -1;
-	//free(ucl->file_name/*UL*/); ucl->file_name/*UL*/ = NULL;
-	return true;
+    if (offset != 0 && lseek/*64*/(ucl->transfer_fd/*UL*/, offset, SEEK_SET) < 0) {
+    	flag_putf(DC_DF_CONNECTIONS, _("%s: Cannot seek in file - %s\n"), quotearg(ucl->local_file/*UL*/), errstr);
+	user_putf(ucl, "$Error File Not Available|");
+	end_upload(ucl, false);
+	return;
     }
-
     ucl->file_size = st.st_size;
+    if (!user_putf(ucl, "$FileLength %llu|", ucl->file_size)) {
+        end_upload(ucl, false);
+        return;
+    }
     ucl->file_pos = offset;
     ucl->transfer_pos = offset;
-
-    /* Allocate slot. This happens after we've opened the file, because if
-     * it would fail, we would have to send back DC_MSG_TRANSFER_FINISHED.
-     */
-    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_STARTING/*UL*/, IPC_INT64, ucl->file_size, IPC_STR, ucl->file_name/*UL*/, IPC_END))
-    	return false;
-    if (!ipc_get(ucl->ipc, DC_MSG_TRANSFER_STARTING/*UL*/, IPC_BOOL, &reply, IPC_END))
-    	return false;
-    if (!reply) {
-    	user_putf(ucl, "$MaxedOut|"); /* Ignore errors, falling through */
-	//transfer_file will be closed at process exit
-	ucl->user_running = false;
-    	return false;
-    }
+    ucl->user_state = DC_USER_SEND_GET;
 
     /* Places to go from here:
      *   user_running becomes false
      *     User process termination. user_disconnect deals with used_*_slots.
      *   $Get received
-     *     Will issue DC_MSG_TRANSFER_FINISHED prior to calling this
+     *     Will issue DC_MSG_UPLOAD_ENDED prior to calling this
      *     function again.
      *   $Send received with file_size == 0
      *   $Send received with file_size != 0, download finished
-     *     Will issue DC_MSG_TRANSFER_FINISHED prior to setting state
+     *     Will issue DC_MSG_UPLOAD_ENDED prior to setting state
      *     to DC_USER_GET.
      */
+}
 
-    if (!user_putf(ucl, "$FileLength %llu|", ucl->file_size))
-	return false;
-
-    ucl->user_state = DC_USER_SEND_GET;
-    return true;
+static void
+upload_file(DCUserConnLocal *ucl)
+{
+    if (ucl->file_size == ucl->file_pos) {
+        /* There's nothing to send - size is 0 or resuming at end. */
+        end_upload(ucl, true);
+        ucl->user_state = DC_USER_GET;
+        return;
+    }
+    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_START, IPC_INT64, ucl->file_pos, IPC_INT64, ucl->file_size, IPC_END)) {
+        end_upload(ucl, true);
+        return;
+    }
+    FD_SET(ucl->user_socket, &ucl->user_write_fds);
+    assert(ucl->user_sendq->cur == 0);
+    ucl->user_state = DC_USER_DATA_SEND;
 }
 
 /* Handle commands and data sent by the hub.
@@ -470,16 +535,12 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
     if (ucl->user_state == DC_USER_DATA_RECV) {
     	ssize_t res;
 
-    	res = full_write(ucl->transfer_file/*DL*/, buf, len);
+    	res = full_write(ucl->transfer_fd/*DL*/, buf, len);
 	if (res < len) {
-	    warn_file_error(res, true, ucl->file_name/*DL*/);
-	    if (close(ucl->transfer_file/*DL*/) < 0)
-	    	warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*DL*/), errstr);
-	    ucl->transfer_file/*DL*/ = -1;
-	    //free(ucl->localfile); ucl->localfile = NULL;
-	    //free(ucl->file_name/*DL*/); ucl->file_name/*DL*/ = NULL;
 	    /* We cannot expect to synchronize with remote at this point, so shut down. */
-	    ucl->user_running = false; //download_next_file(ucl);
+	    warn_file_error(res, true, ucl->share_file/*DL*/);
+	    end_download(ucl, false, "local error");
+	    ucl->user_running = false; 
 	    return;
 	}
 	ucl->file_pos += len;
@@ -487,22 +548,8 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	ucl->data_size/*DL*/ -= len;
     	send_user_status(ucl, ucl->file_pos);
     	if (ucl->file_pos == ucl->file_size) {
-	    flag_putf(DC_DF_DOWNLOAD, _("%s: Download complete.\n"), quotearg(base_name(ucl->file_name/*DL*/)));/* XXX: maybe move main.c DC_MSG_TRANSFER_FINISHED?*/
-	    if (close(ucl->transfer_file/*DL*/) < 0)
-	    	warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*DL*/), errstr);
-            if (ucl->file_flag == DC_TF_NORMAL) {
-                char *tmp;
-                tmp = xstrndup(ucl->localfile, strlen(ucl->localfile)-5);
-                if (rename(ucl->localfile, tmp) < 0)
-                    warn(_("%s: Cannot rename file to `%s' - %s\n"), quotearg_n(0, ucl->localfile), quotearg_n(1, tmp), errstr);
-                free(tmp);
-            }
-            //free(ucl->localfile); ucl->localfile = NULL;
-	    //free(ucl->file_name/*DL*/); ucl->file_name/*DL*/ = NULL; ucl->base_name = NULL;
-	    ucl->transfer_file/*DL*/ = -1;
-	    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_FINISHED/*DL*/, IPC_INT32, ucl->file_flag, IPC_END))
-	    	return;
-	    download_next_file(ucl); /* Will change state for us. */
+	    end_download(ucl, true, "transfer complete");
+	    download_next_file(ucl);
 	    return;
 	}
     }
@@ -592,23 +639,18 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	if (!wants_to_download(ucl, &we_download))
 	    return;
 
-    	if (they_download) {
-	    // Remote wants to download. Do we want to download too?
+    	if (they_download) { /* Remote wants to download. Do we want to download too? */
 	    if (we_download) {
-		if (remote_rand >= ucl->dir_rand) {
-		    // We lost. Let them download.
+		if (remote_rand >= ucl->dir_rand) { /* We lost. Let them download. */
 		    ucl->our_dir = DC_DIR_SEND;
-		    /* XXX: what should happen if remote_rand == dir_rand!? */
-		} else {
-		    // We won.
+		} else { /* We won. */
 		    ucl->our_dir = DC_DIR_RECEIVE;
 		}
-	    } else {
-		// We don't want to download anything. Let remote download.
+                /* XXX: what should happen if remote_rand == dir_rand!? */
+	    } else { /* We don't want to download anything. Let remote download. */
 		ucl->our_dir = DC_DIR_SEND;
 	    }
-	} else {
-	    // Remote wants to upload. Do we want to upload too?
+	} else { /* Remote wants to upload. Do we want to upload too? */
 	    if (!we_download) {
 		warn(_("User does not want to download, nor do we.\n"));
 		ucl->user_running = false;
@@ -659,44 +701,38 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
 	    	offset--;
 	}
 
-	if (ucl->user_state == DC_USER_SEND_GET) {
-	    /* XXX: this probably meant that the remote user
-	     * decided he/she didn't want the file (somehow).
-	     */
-	    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_FINISHED/*UL*/, IPC_INT32, 0, IPC_END))
-	    	return;
-	    if (close(ucl->transfer_file/*UL*/) < 0)
-		warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-	    ucl->transfer_file/*UL*/ = -1;
-	}
+        /* Maybe the remote user did not want the file after all? */
+	if (ucl->user_state == DC_USER_SEND_GET)
+	    end_upload(ucl, false);
+
     	open_upload_file(ucl, buf+5, offset); /* Changes state */
     }
     else if (len == 9 && strcmp(buf, "$MaxedOut") == 0) {
     	if (!check_state(ucl, buf, DC_USER_FILE_LENGTH))
 	    return;
-	warn(_("User is maxed out.\n")); /*XXX: or DC_DF_DOWNLOAD?*/
+	end_download(ucl, false, "remote is maxed out");
 	ucl->user_running = false;
     }
     else if (len >= 12 && strncmp(buf, "$FileLength ", 12) == 0) {
+        uint64_t file_size;
+
     	if (!check_state(ucl, buf, DC_USER_FILE_LENGTH))
 	    return;
-	if (!parse_uint64(buf+12, &ucl->file_size)) {
-	    warn(_("Invalid $FileLength message: Invalid length integer\n"));
-	    ucl->user_running = false;
+	if (!parse_uint64(buf+12, &file_size)) {
+	    end_download(ucl, false, "protocol error: invalid $FileLength message");
+	    download_next_file(ucl);
 	    return;
 	}
-    	open_download_file(ucl); /* Will change state */
+    	open_download_file(ucl, file_size); /* Will change state */
     }
     else if (len >= 7 && strncmp(buf, "$Error ", 7) == 0) {
 	if (ucl->user_state == DC_USER_FILE_LENGTH) {
     	    if (strcmp(buf+7, "File Not Available") == 0) {
-	    	warn("%s: File not available on remote\n", quotearg(ucl->file_name/*DL*/));
-		if (close(ucl->transfer_file/*DL*/) < 0)
-	    	    warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*DL*/), errstr);
-		// free(ucl->file_name/*DL*/); ucl->file_name/*UL*/ = NULL;
-		ucl->transfer_file/*DL*/ = -1;
-		download_next_file(ucl);
+	    	end_download(ucl, false, "file not available on remote");
+	    } else {
+	        end_download(ucl, false, "remote error: %s", quotearg(buf+7));
 	    }
+            download_next_file(ucl);
 	    return;
 	}
 	warn(_("Received error from user: %s\n"), quotearg(buf+7));
@@ -705,21 +741,11 @@ user_handle_command(DCUserConnLocal *ucl, char *buf, uint32_t len)
     else if (len >= 5 && strncmp(buf, "$Send ", 5) == 0) {
     	if (!check_state(ucl, buf, DC_USER_SEND_GET))
 	    return;
-	if (ucl->file_size != ucl->file_pos) { /* size is 0 or starting at end */
-    	    FD_SET(ucl->user_socket, &ucl->user_write_fds);
-	    assert(ucl->user_sendq->cur == 0);
-	    ucl->user_state = DC_USER_DATA_SEND;
-	} else {
-	    /* XXX: this probably meant that the remote user
-	     * decided he/she didn't want the file (somehow).
-	     */
-	    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_FINISHED/*UL*/, IPC_INT32, 0, IPC_END))
-	    	return;
-	    ucl->user_state = DC_USER_GET;
-	}
+        upload_file(ucl);
     }
 }
 
+
 static void
 user_input_available(DCUserConnLocal *ucl)
 {
@@ -808,13 +834,10 @@ user_now_writable(DCUserConnLocal *ucl)
     	assert(ucl->file_size != 0);
     	block = MIN(DEFAULT_SENDQ_SIZE, ucl->file_size - ucl->file_pos);
 	if (block > 0 && ucl->user_sendq->cur == 0) { //if (ucl->user_sendq->cur < ucl->user_sendq->max) {
-	    res = byteq_full_read_upto(ucl->user_sendq, ucl->transfer_file/*UL*/, block);
+	    res = byteq_full_read_upto(ucl->user_sendq, ucl->transfer_fd/*UL*/, block);
 	    if (res < block) {
-	    	warn_file_error(res, false, ucl->file_name/*UL*/);
-		if (close(ucl->transfer_file/*UL*/) < 0)
-		    warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-		ucl->transfer_file/*UL*/ = -1;
-		//free(ucl->file_name/*UL*/); ucl->file_name/*UL*/ = NULL;
+	    	warn_file_error(res, false, ucl->local_file/*UL*/);
+	    	end_upload(ucl, false);
 		ucl->user_running = false;
 		return;
 	    }
@@ -825,6 +848,7 @@ user_now_writable(DCUserConnLocal *ucl)
 	res = byteq_write(ucl->user_sendq, ucl->user_socket);
 	if (res == 0 || (res < 0 && errno != EAGAIN)) {
     	    warn_socket_error(res, true, _("user"));
+            end_upload(ucl, false);
 	    ucl->user_running = false;
 	    return;
 	}
@@ -832,14 +856,8 @@ user_now_writable(DCUserConnLocal *ucl)
   	send_user_status(ucl, ucl->file_pos - ucl->user_sendq->cur);
 
 	if (ucl->file_pos == ucl->file_size && ucl->user_sendq->cur == 0) {
-	    flag_putf(DC_DF_UPLOAD, _("%s: Upload complete.\n"), base_name(ucl->file_name/*UL*/));/*XXX: maybe move main.c DC_MSG_TRANSFER_FINISHED? */
     	    FD_CLR(ucl->user_socket, &ucl->user_write_fds);
-	    if (close(ucl->transfer_file/*UL*/) < 0)
-	    	warn(_("%s: Cannot close file - %s\n"), quotearg(ucl->file_name/*UL*/), errstr);
-	    ucl->transfer_file = -1;
-    	    free(ucl->file_name/*UL*/); ucl->file_name/*UL*/ = NULL;
-	    if (!ipc_put(ucl->ipc, DC_MSG_TRANSFER_FINISHED/*UL*/, IPC_INT32, 0, IPC_END))
-	    	return;
+    	    end_upload(ucl, true);
 	    ucl->user_state = DC_USER_GET;
 	}
     }
@@ -900,7 +918,7 @@ read_signal_input(DCUserConnLocal *ucl)
     }
 }
 
-static void
+void
 __attribute__((noreturn))
 user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 {
@@ -910,12 +928,11 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
     ucl = xmalloc(sizeof(DCUserConnLocal));
     cur_ucl = ucl; /* one per process anyway */
     ucl->user_nick = NULL;
-    ucl->base_path = NULL;
-    ucl->file_name = NULL;
-    ucl->localfile = NULL;
+    ucl->share_file = NULL;
+    ucl->local_file = NULL;
     ucl->user_recvq_last = 0;
     ucl->user_socket = -1;
-    ucl->transfer_file = -1;
+    ucl->transfer_fd = -1;
     ucl->data_size = 0;     /* only useful when receiving files */
     ucl->file_pos = 0;
     ucl->transfer_pos = 0;
@@ -955,14 +972,9 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 	goto cleanup;
     }
 
-    //ucl->user_recvq_last = 0;
     ucl->user_recvq = byteq_new(DEFAULT_RECVQ_SIZE);
     ucl->user_sendq = byteq_new(DEFAULT_SENDQ_SIZE);
-    //ucl->user_state = DC_USER_MYNICK;*/
     ucl->dir_rand = rand() % 0x8000;
-    //ucl->file_pos = 0;
-    //ucl->file_size = 0;
-    //ucl->data_size = 0;
     if (sock < 0) {
 	ucl->user_socket = socket(PF_INET, SOCK_STREAM, 0);
 	if (ucl->user_socket < 0) {
@@ -1010,9 +1022,10 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 	    break;
 	}
 
-    	/* Must check for user writable prior to readable, to detect
+    	/* We must check for user writable prior to readable, to detect
 	 * connect() completion before any data is received and handled.
 	 */
+
     	if (ucl->user_running && FD_ISSET(ucl->signal_pipe[0], &res_read_fds))
 	    read_signal_input(ucl);
 	if (ucl->user_running && FD_ISSET(ucl->main_socket, &res_read_fds))
@@ -1026,150 +1039,22 @@ user_main(int main_sockets[2], struct sockaddr_in *addr, int sock)
 cleanup:
 
     ipc_free(ucl->ipc);
-    free(ucl->localfile);
-    free(ucl->file_name);
+    free(ucl->local_file);
+    free(ucl->share_file);
     free(ucl->user_nick);
     byteq_free(ucl->user_recvq);
     byteq_free(ucl->user_sendq);
 
-    if (ucl->transfer_file >= 0 && close(ucl->transfer_file) < 0)
+    if (ucl->transfer_fd >= 0 && close(ucl->transfer_fd) < 0)
     	warn(_("Cannot close transfer file - %s\n"), errstr); /* XXX: should print WHAT file - then remove " transfer" */
     if (ucl->signal_pipe[0] >= 0 && close(ucl->signal_pipe[0]) < 0)
     	warn(_("Cannot close signal pipe - %s\n"), errstr);
     if (ucl->signal_pipe[1] >= 0 && close(ucl->signal_pipe[1]) < 0)
     	warn(_("Cannot close signal pipe - %s\n"), errstr);
     if (ucl->main_socket >= 0 && close(ucl->main_socket) < 0)
-    	warn(_("Cannot close main process socket - %s\n"), errstr);	/* XXX: "socket" or "connection" */
+    	warn(_("Cannot close main process socket - %s\n"), errstr); /* XXX: "socket" or "connection" */
     if (ucl->user_socket >= 0 && close(ucl->user_socket) < 0)
     	warn(_("Cannot close user connection - %s\n"), errstr);
 
     exit(EXIT_SUCCESS);
 }
-
-DCUserConn *
-user_connection_new(struct sockaddr_in *addr, int user_socket)
-{
-    DCUserConn *uc;
-    int main_sockets[2];
-    pid_t pid;
-
-    if (socketpair(AF_LOCAL, SOCK_STREAM, 0, main_sockets) < 0) {
-    	warn(_("Cannot create socket pair - %s\n"), errstr);
-	return NULL;
-    }
-
-    pid = fork();
-    if (pid < 0) {
-    	warn(_("Cannot create process - %s\n"), errstr);
-	if (close(main_sockets[0]) < 0)
-	    warn(_("Cannot close socket - %s\n"), errstr);
-	if (close(main_sockets[1]) < 0)
-	    warn(_("Cannot close socket - %s\n"), errstr);
-	return NULL;
-    }
-    if (pid == 0)
-	user_main(main_sockets, addr, user_socket);
-
-    if (close(main_sockets[1]) < 0)
-    	warn(_("Cannot close socket - %s\n"), errstr);
-    if (user_socket >= 0 && close(user_socket) < 0)
-    	warn(_("Cannot close socket - %s\n"), errstr);
-
-    uc = xmalloc(sizeof(DCUserConn));
-    uc->pid = pid;
-    uc->main_socket = main_sockets[0];
-    uc->info = NULL;
-    uc->occupied_slot = 0;
-    uc->dir = DC_DIR_UNKNOWN;
-    uc->transfer_file = NULL;
-    //uc->we_connected = (user_socket < 0);
-    add_user_conn(uc);
-    FD_SET(uc->main_socket, &read_fds);
-
-    return uc;
-}
-
-void
-user_disconnect(DCUserConn *uc)
-{
-    /*int status;*/
-
-    flag_putf(DC_DF_CONNECTIONS, _("User %s: Shutting down user connection.\n"), quotearg(uc->name)); /* XXX: move where? */
-    ipc_free(uc->ipc);
-    if (uc->info != NULL) {
-    	int c;
-	for (c = 0; c < uc->info->conn_count; c++) {
-	    if (uc->info->conn[c] == uc)
-		break;
-	}
-	assert(c < uc->info->conn_count);
-	uc->info->conn[c] = NULL;
-    	uc->info->conn_count--;
-	for (; c < uc->info->conn_count; c++)
-	    uc->info->conn[c] = uc->info->conn[c+1];
-	user_info_free(uc->info);
-    }
-    if (uc->occupied_slot) {
-    	if (uc->dir == DC_DIR_SEND) {
-	    used_ul_slots--;
-	} else { /* uc->dir must be DC_DIR_RECV */
-	    used_dl_slots--;
-	}
-    }
-    FD_CLR(uc->main_socket, &read_fds);
-    FD_CLR(uc->main_socket, &write_fds);
-    if (close(uc->main_socket) < 0)
-	warn(_("Cannot close socket - %s\n"), errstr);
-    free(uc->transfer_file);
-    /*if (waitpid(uc->pid, &status, 0) < 0)
-    	warn(_("Cannot wait for process - %s\n"), errstr);
-    else if (!WIFEXITED(status))
-    	warn(_("User process exited with status %d\n"), status);*/
-    free(uc);
-}
-
-char *
-user_conn_status_to_string(DCUserConn *uc, time_t now)
-{
-    char *status;
-
-    if (uc->occupied_slot) {
-    	int percent = (uc->transfer_pos * 100)/MAX(uc->transfer_len, 1);
-	int rate = (now == 0 ? 0 : uc->transfer_pos/1024/MAX(now-uc->transfer_start, 1));
-	char *file = base_name(uc->transfer_file);
-	status = xasprintf(
-	    uc->dir == DC_DIR_RECEIVE
-		? _("Downloading %3d%% (at %5d kb/s) %s")
-		: _("Uploading   %3d%% (at %5d kb/s) %s"),
-	    percent, rate, file);
-    } else {
-	status = xasprintf(_("Idle"));
-    }
-
-    return status;
-}
-
-bool
-parse_ip_and_port(char *source, struct sockaddr_in *addr, uint16_t defport)
-{
-    char *port;
-
-    port = strchr(source, ':');
-    if (port == NULL) {
-    	if (defport == 0)
-	    return false;
-	addr->sin_port = htons(defport);
-    } else {
-	*port = '\0';
-	port++;
-	if (!parse_uint16(port, &addr->sin_port))
-    	    return false;
-        addr->sin_port = htons(addr->sin_port);
-    }
-
-    if (!inet_aton(source, &addr->sin_addr))
-    	return false;
-
-    addr->sin_family = AF_INET;
-    return true;
-}
