@@ -24,6 +24,7 @@
 #include <unistd.h>		/* POSIX: confstr, ... */
 #include <fcntl.h>		/* ? */
 #include <string.h>		/* C89 */
+#include <arpa/inet.h>		/* ? */
 #include "xalloc.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
@@ -32,8 +33,50 @@
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/comparison.h"
+#include "common/intutil.h"
 #include "microdc.h"
 
+/* safe_rename: A rename that won't overwrite the destination.
+ * Or at least try not to. :)
+ */
+int
+safe_rename(const char *oldpath, const char *newpath)
+{
+    int fd;
+
+    fd = open(newpath, O_CREAT|O_EXCL, 0600);
+    if (fd < 0)
+        return -1;
+    if (close(fd) != 0)
+        return -1;
+    return rename(oldpath, newpath);
+}
+
+bool
+parse_ip_and_port(char *source, struct sockaddr_in *addr, uint16_t defport)
+{
+    char *port;
+
+    port = strrchr(source, ':');
+    if (port == NULL) {
+        if (defport == 0)
+            return false;
+        addr->sin_port = htons(defport);
+    } else {
+        *port = '\0';
+        port++;
+        if (!parse_uint16(port, &addr->sin_port))
+            return false;
+        addr->sin_port = htons(addr->sin_port);
+    }
+
+    if (!inet_aton(source, &addr->sin_addr))
+        return false;
+
+    addr->sin_family = AF_INET;
+    return true;
+}
+
 /* Move to lib !? */
 int
 ilog10(uint64_t c)
@@ -210,25 +253,6 @@ wordwrap(const char *str, size_t len, size_t first_width, size_t other_width)
     return out;
 }
 
-/* for Gnulib? */
-#if 0
-char *
-xconfstr(const char *name)
-{
-    char *buf;
-    size_t len;
-
-    len = confstr(name, NULL, 0);
-    if (len != 0) {
-	buf = xmalloc(len);
-	if (confstr(name, buf, len) == 0)
-	    return buf;
-	free(buf);
-    }
-    return NULL;
-}
-#endif
-
 char *
 join_strings(char **strs, int count, char mid)
 {
