@@ -89,6 +89,25 @@ typedef struct {
     const char *name;
 } DCDisplayFlagDetails;
 
+/* settings/variables */
+int used_ul_slots = 0;
+int used_dl_slots = 0;
+uint32_t my_ul_slots;
+char *my_nick;
+char *my_tag;
+char *my_description;
+char *my_speed;
+char *my_email;
+char *share_dir = NULL;
+char *download_dir;
+char *listing_dir;
+char *my_password;
+bool is_active;
+uint64_t my_share_size = 0;
+uint32_t display_flags = ~(DC_DF_DEBUG); /* All flags except debug set */
+uint32_t log_flags = ~(DC_DF_DEBUG);
+struct in_addr force_listen_addr = { INADDR_NONE };
+
 /* This list must be sorted according to strcmp. */
 DCDisplayFlagDetails display_flag_details[] =  {
     { DC_DF_CONNECTIONS, 	"connections" },
@@ -102,16 +121,16 @@ DCDisplayFlagDetails display_flag_details[] =  {
 static const int display_flag_count = sizeof(display_flag_details)/sizeof(*display_flag_details);
 
 static char *speeds[] = {
-    "Modem",
     "28.8Kbps",
     "33.6Kbps",
     "56Kbps",
-    "Satellite",
-    "ISDN",
-    "DSL",
     "Cable",
+    "DSL",
+    "ISDN",
     "LAN(T1)",
     "LAN(T3)",
+    "Modem",
+    "Satellite",
 };
 static const int speeds_count = sizeof(speeds)/sizeof(*speeds);
 
@@ -444,6 +463,7 @@ bool_completion_generator(DCCompletionInfo *ci)
     }
     if (strleftcmp(ci->word, p0) == 0)
         ptrv_append(ci->results, new_completion_entry(xstrdup(p0), NULL));
+    ptrv_sort(ci->results, completion_entry_display_compare);
 }
 
 static bool
